from __future__ import annotations

from typing import Any
from uuid import uuid4

import pydantic
from prefect import flow, get_run_logger
from prefect.client.schemas.objects import State  # noqa: TC002
from prefect.states import Completed, Failed

from infrahub import lock
from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.branch.enums import BranchStatus
from infrahub.core.changelog.diff import DiffChangelogCollector, MigrationTracker
from infrahub.core.constants import MutationAction
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.diff_locker import DiffLocker
from infrahub.core.diff.ipam_diff_parser import IpamDiffParser
from infrahub.core.diff.merger.merger import DiffMerger
from infrahub.core.diff.model.path import BranchTrackingId, EnrichedDiffRoot, EnrichedDiffRootMetadata
from infrahub.core.diff.models import RequestDiffUpdate
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.graph import GRAPH_VERSION
from infrahub.core.merge import BranchMerger
from infrahub.core.migrations.exceptions import MigrationFailureError
from infrahub.core.migrations.runner import MigrationRunner
from infrahub.core.migrations.schema.models import SchemaApplyMigrationData
from infrahub.core.migrations.schema.tasks import schema_apply_migrations
from infrahub.core.timestamp import Timestamp
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.events.branch_action import (
    BranchCreatedEvent,
    BranchDeletedEvent,
    BranchMergedEvent,
    BranchMigratedEvent,
    BranchRebasedEvent,
)
from infrahub.events.models import EventMeta, InfrahubEvent
from infrahub.events.node_action import get_node_event
from infrahub.exceptions import BranchNotFoundError, ValidationError
from infrahub.generators.constants import GeneratorDefinitionRunSource
from infrahub.graphql.mutations.models import BranchCreateModel  # noqa: TC001
from infrahub.workers.dependencies import get_component, get_database, get_event_service, get_workflow
from infrahub.workflows.catalogue import (
    BRANCH_CANCEL_PROPOSED_CHANGES,
    BRANCH_MERGE_POST_PROCESS,
    DIFF_REFRESH_ALL,
    DIFF_UPDATE,
    GIT_REPOSITORIES_CREATE_BRANCH,
    IPAM_RECONCILIATION,
    TRIGGER_ARTIFACT_DEFINITION_GENERATE,
    TRIGGER_GENERATOR_DEFINITION_RUN,
)
from infrahub.workflows.utils import add_tags


@flow(name="branch-migrate", flow_run_name="Apply migrations to branch {branch}")
async def migrate_branch(branch: str, context: InfrahubContext, send_events: bool = True) -> None:
    await add_tags(branches=[branch])

    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()

        obj = await Branch.get_by_name(db=db, name=branch)

        if obj.graph_version == GRAPH_VERSION:
            log.info(f"Branch '{obj.name}' has graph version {obj.graph_version}, no migrations to apply")
            return

        migration_runner = MigrationRunner(branch=obj)
        if not migration_runner.has_migrations():
            log.info(f"No migrations detected for branch '{obj.name}'")
            obj.graph_version = GRAPH_VERSION
            await obj.save(db=db)
            return

        # Branch status will remain as so if the migration process fails
        # This will help user to know that a branch is in an invalid state to be used properly and that actions need to be taken
        if obj.status != BranchStatus.NEED_UPGRADE_REBASE:
            obj.status = BranchStatus.NEED_UPGRADE_REBASE
            await obj.save(db=db)

        try:
            log.info(f"Running migrations for branch '{obj.name}'")
            await migration_runner.run(db=db, at=Timestamp())
        except MigrationFailureError as exc:
            log.error(f"Failed to run migrations for branch '{obj.name}': {exc.errors}")
            raise

        if obj.status == BranchStatus.NEED_UPGRADE_REBASE:
            obj.status = BranchStatus.OPEN
        obj.graph_version = GRAPH_VERSION
        await obj.save(db=db)

    if send_events:
        event_service = await get_event_service()
        await event_service.send(
            BranchMigratedEvent(
                branch_name=obj.name, branch_id=str(obj.uuid), meta=EventMeta(branch=obj, context=context)
            )
        )


@flow(name="branch-rebase", flow_run_name="Rebase branch {branch}")
async def rebase_branch(branch: str, context: InfrahubContext, send_events: bool = True) -> None:  # noqa: PLR0915
    workflow = get_workflow()
    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()
        await add_tags(branches=[branch])
        obj = await Branch.get_by_name(db=db, name=branch)
        base_branch = await Branch.get_by_name(db=db, name=registry.default_branch)
        component_registry = get_component_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
        diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
        initial_from_time = Timestamp(obj.get_branched_from())
        merger = BranchMerger(
            db=db,
            diff_coordinator=diff_coordinator,
            diff_merger=diff_merger,
            diff_repository=diff_repository,
            source_branch=obj,
            diff_locker=DiffLocker(),
            workflow=workflow,
        )

        enriched_diff_metadata = await diff_coordinator.update_branch_diff(base_branch=base_branch, diff_branch=obj)
        async for _ in diff_repository.get_all_conflicts_for_diff(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        ):
            # if there are any conflicts, raise the error
            raise ValidationError(
                f"Branch {obj.name} contains conflicts with the default branch that must be addressed."
                " Please review the diff for details and manually update the conflicts before rebasing."
            )
        node_diff_field_summaries = await diff_repository.get_node_field_summaries(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        )

        candidate_schema = merger.get_candidate_schema()
        determiner = ConstraintValidatorDeterminer(schema_branch=candidate_schema)
        constraints = await determiner.get_constraints(node_diffs=node_diff_field_summaries)

        # If there are some changes related to the schema between this branch and main, we need to
        #  - Run all the validations to ensure everything is correct before rebasing the branch
        #  - Run all the migrations after the rebase
        if obj.has_schema_changes:
            constraints += await merger.calculate_validations(target_schema=candidate_schema)
        if constraints:
            responses = await schema_validate_migrations(
                message=SchemaValidateMigrationData(branch=obj, schema_branch=candidate_schema, constraints=constraints)
            )
            error_messages = [
                f"{violation.message} for constraint {response.constraint_name} {response.schema_path.field_name} {response.schema_path.property_name} and node {violation.node_id} {violation.node_kind}"  # noqa: E501
                for response in responses
                for violation in response.violations
            ]

            if error_messages:
                raise ValidationError(",\n".join(error_messages))

        schema_in_main_before = merger.destination_schema.duplicate()
        migrations = []
        async with lock.registry.global_graph_lock():
            async with db.start_transaction() as dbt:
                rebase_at = Timestamp()
                await obj.rebase(db=dbt, user_id=context.account.account_id, at=rebase_at)
                log.info("Branch successfully rebased")

            if obj.has_schema_changes:
                # NOTE there is a bit additional work in order to calculate a proper diff that will
                # allow us to pull only the part of the schema that has changed, for now the safest option is to pull
                # Everything
                # schema_diff = await merger.has_schema_changes()
                # TODO Would be good to convert this part to a Prefect Task in order to track it properly
                updated_schema = await registry.schema.load_schema_from_db(
                    db=db,
                    branch=obj,
                    # schema=merger.source_schema.duplicate(),
                    # schema_diff=schema_diff,
                )
                registry.schema.set_schema_branch(name=obj.name, schema=updated_schema)
                obj.update_schema_hash()
                await obj.save(db=db, user_id=context.account.account_id)

                # Execute the migrations
                migrations = await merger.calculate_migrations(target_schema=updated_schema)

                errors = await schema_apply_migrations(
                    message=SchemaApplyMigrationData(
                        branch=merger.source_branch,
                        new_schema=candidate_schema,
                        previous_schema=schema_in_main_before,
                        migrations=migrations,
                        user_id=context.account.account_id,
                        at=rebase_at,
                    )
                )
                for error in errors:
                    log.error(error)

        default_branch_diff = await _get_diff_root(
            diff_coordinator=diff_coordinator,
            enriched_diff_metadata=enriched_diff_metadata,
            diff_repository=diff_repository,
            base_branch=base_branch,
            target_from=initial_from_time,
        )

        # -------------------------------------------------------------
        # Trigger the reconciliation of IPAM data after the rebase
        # -------------------------------------------------------------
        diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
        ipam_node_details = await diff_parser.get_changed_ipam_node_details(
            source_branch_name=obj.name,
            target_branch_name=registry.default_branch,
        )
        if ipam_node_details:
            await workflow.submit_workflow(
                workflow=IPAM_RECONCILIATION,
                context=context,
                parameters={"branch": obj.name, "ipam_node_details": ipam_node_details},
            )

    await migrate_branch(branch=branch, context=context, send_events=send_events)
    await workflow.submit_workflow(workflow=DIFF_REFRESH_ALL, context=context, parameters={"branch_name": obj.name})

    if not send_events:
        return

    # -------------------------------------------------------------
    # Generate an event to indicate that a branch has been rebased
    # -------------------------------------------------------------
    rebase_event = BranchRebasedEvent(
        branch_name=obj.name, branch_id=str(obj.uuid), meta=EventMeta(branch=obj, context=context)
    )
    events: list[InfrahubEvent] = [rebase_event]
    changelog_collector = DiffChangelogCollector(
        diff=default_branch_diff, branch=obj, db=db, migration_tracker=MigrationTracker(migrations=migrations)
    )
    for action, node_changelog in changelog_collector.collect_changelogs():
        node_event_class = get_node_event(MutationAction.from_diff_action(diff_action=action))
        mutate_event = node_event_class(
            kind=node_changelog.node_kind,
            node_id=node_changelog.node_id,
            changelog=node_changelog,
            fields=node_changelog.updated_fields,
            meta=EventMeta.from_parent(parent=rebase_event, branch=obj),
        )
        events.append(mutate_event)

    event_service = await get_event_service()
    for event in events:
        await event_service.send(event)


@flow(name="branch-merge", flow_run_name="Merge branch {branch} into main")
async def merge_branch(branch: str, context: InfrahubContext, proposed_change_id: str | None = None) -> None:
    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()

        await add_tags(branches=[branch, registry.default_branch])

        obj = await Branch.get_by_name(db=db, name=branch)
        default_branch = await registry.get_branch(db=db, branch=registry.default_branch)
        component_registry = get_component_registry()
        merge_event = BranchMergedEvent(
            branch_name=obj.name,
            branch_id=str(obj.get_uuid()),
            proposed_change_id=proposed_change_id,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )

        merger: BranchMerger | None = None
        async with lock.registry.global_graph_lock():
            diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
            diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
            diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
            merger = BranchMerger(
                db=db,
                diff_coordinator=diff_coordinator,
                diff_merger=diff_merger,
                diff_repository=diff_repository,
                source_branch=obj,
                diff_locker=DiffLocker(),
                workflow=get_workflow(),
            )
            merge_at = Timestamp()
            branch_diff = await merger.merge(at=merge_at)
            await merger.update_schema()

        changelog_collector = DiffChangelogCollector(diff=branch_diff, branch=obj, db=db)
        node_events = changelog_collector.collect_changelogs()
        if merger and merger.migrations:
            errors = await schema_apply_migrations(
                message=SchemaApplyMigrationData(
                    branch=merger.destination_branch,
                    new_schema=merger.destination_schema,
                    previous_schema=merger.initial_source_schema,
                    migrations=merger.migrations,
                    user_id=context.account.account_id,
                    at=merge_at,
                )
            )
            for error in errors:
                log.error(error)

        # -------------------------------------------------------------
        # Trigger the reconciliation of IPAM data after the merge
        # -------------------------------------------------------------
        diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
        ipam_node_details = await diff_parser.get_changed_ipam_node_details(
            source_branch_name=obj.name,
            target_branch_name=registry.default_branch,
        )
        if ipam_node_details:
            await get_workflow().submit_workflow(
                workflow=IPAM_RECONCILIATION,
                context=context,
                parameters={"branch": registry.default_branch, "ipam_node_details": ipam_node_details},
            )
        # -------------------------------------------------------------
        # remove tracking ID from the diff because there is no diff after the merge
        # -------------------------------------------------------------
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        await diff_repository.mark_tracking_ids_merged(tracking_ids=[BranchTrackingId(name=obj.name)])

        # -------------------------------------------------------------
        # Generate an event to indicate that a branch has been merged
        # NOTE: we still need to convert this event and potentially pull
        #   some tasks currently executed based on the event into this workflow
        # -------------------------------------------------------------
        await get_workflow().submit_workflow(
            workflow=BRANCH_MERGE_POST_PROCESS,
            context=context,
            parameters={"source_branch": obj.name, "target_branch": registry.default_branch},
        )

        events: list[InfrahubEvent] = [merge_event]

        for action, node_changelog in node_events:
            meta = EventMeta.from_parent(parent=merge_event, branch=default_branch)
            node_event_class = get_node_event(MutationAction.from_diff_action(diff_action=action))
            mutate_event = node_event_class(
                kind=node_changelog.node_kind,
                node_id=node_changelog.node_id,
                changelog=node_changelog,
                fields=node_changelog.updated_fields,
                meta=meta,
            )
            events.append(mutate_event)

        event_service = await get_event_service()
        for event in events:
            await event_service.send(event=event)


@flow(name="branch-delete", flow_run_name="Delete branch {branch}")
async def delete_branch(branch: str, context: InfrahubContext) -> None:
    await add_tags(branches=[branch])

    database = await get_database()
    async with database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=str(branch))
        await obj.delete(db=db)

        event = BranchDeletedEvent(
            branch_name=branch,
            branch_id=str(obj.uuid),
            sync_with_git=obj.sync_with_git,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )

        await get_workflow().submit_workflow(
            workflow=BRANCH_CANCEL_PROPOSED_CHANGES, context=context, parameters={"branch_name": branch}
        )

        event_service = await get_event_service()
        await event_service.send(event=event)


@flow(
    name="branch-validate",
    flow_run_name="Validate branch {branch} for conflicts",
    description="Validate if the branch has some conflicts",
    persist_result=True,
)
async def validate_branch(branch: str) -> State:
    await add_tags(branches=[branch])

    database = await get_database()
    async with database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=branch)

        component_registry = get_component_registry()
        diff_repo = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        has_conflicts = await diff_repo.diff_has_conflicts(
            diff_branch_name=obj.name, tracking_id=BranchTrackingId(name=obj.name)
        )
        if has_conflicts:
            return Failed(message="branch has some conflicts")
        return Completed(message="branch is valid")


@flow(name="create-branch", flow_run_name="Create branch {model.name}")
async def create_branch(model: BranchCreateModel, context: InfrahubContext) -> None:
    await add_tags(branches=[model.name])

    database = await get_database()
    async with database.start_session() as db:
        try:
            await Branch.get_by_name(db=db, name=model.name)
            raise ValidationError(f"The branch {model.name} already exists")
        except BranchNotFoundError:
            pass

        data_dict: dict[str, Any] = dict(model)
        data_dict.pop("is_isolated", None)

        try:
            obj = Branch(**data_dict)
        except pydantic.ValidationError as exc:
            error_msgs = [f"invalid field {error['loc'][0]}: {error['msg']}" for error in exc.errors()]
            raise ValidationError("\n".join(error_msgs)) from exc

        async with lock.registry.local_schema_lock():
            # Copy the schema from the origin branch and set the hash and the schema_changed_at value
            origin_schema = registry.schema.get_schema_branch(name=obj.origin_branch)
            new_schema = origin_schema.duplicate(name=obj.name)
            registry.schema.set_schema_branch(name=obj.name, schema=new_schema)
            obj.update_schema_hash()
            await obj.save(db=db, user_id=context.account.account_id)

            # Add Branch to registry
            registry.branch[obj.name] = obj
            component = await get_component()
            await component.refresh_schema_hash(branches=[obj.name])

        event = BranchCreatedEvent(
            branch_name=obj.name,
            branch_id=str(obj.uuid),
            sync_with_git=obj.sync_with_git,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )
        event_service = await get_event_service()
        await event_service.send(event=event)

        if obj.sync_with_git:
            await get_workflow().submit_workflow(
                workflow=GIT_REPOSITORIES_CREATE_BRANCH,
                context=context,
                parameters={"branch": obj.name, "branch_id": str(obj.uuid)},
            )


async def _get_diff_root(
    diff_coordinator: DiffCoordinator,
    enriched_diff_metadata: EnrichedDiffRootMetadata,
    diff_repository: DiffRepository,
    base_branch: Branch,
    target_from: Timestamp,
) -> EnrichedDiffRoot:
    default_branch_diff = await diff_coordinator.create_or_update_arbitrary_timeframe_diff(
        base_branch=base_branch,
        diff_branch=base_branch,
        from_time=target_from,
        to_time=enriched_diff_metadata.to_time,
        name=str(uuid4()),
    )
    # make sure we have the actual diff with data and not just the metadata
    if not isinstance(default_branch_diff, EnrichedDiffRoot):
        default_branch_diff = await diff_repository.get_one(
            diff_branch_name=base_branch.name, diff_id=default_branch_diff.uuid
        )

    return default_branch_diff


@flow(
    name="branch-merge-post-process",
    flow_run_name="Run additional tasks after merging {source_branch} in {target_branch}",
)
async def post_process_branch_merge(source_branch: str, target_branch: str, context: InfrahubContext) -> None:
    database = await get_database()
    async with database.start_session() as db:
        await add_tags(branches=[source_branch])
        log = get_run_logger()
        log.info(f"Running additional tasks after merging {source_branch} within {target_branch}")

        component_registry = get_component_registry()
        default_branch = registry.get_branch_from_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=default_branch)
        # send diff update requests for every branch-tracking diff
        branch_diff_roots = await diff_repository.get_roots_metadata(base_branch_names=[target_branch])

        await get_workflow().submit_workflow(
            workflow=TRIGGER_ARTIFACT_DEFINITION_GENERATE,
            context=context,
            parameters={"branch": target_branch},
        )

        await get_workflow().submit_workflow(
            workflow=TRIGGER_GENERATOR_DEFINITION_RUN,
            context=context,
            parameters={"branch": target_branch, "source": GeneratorDefinitionRunSource.MERGE},
        )

        active_branches = await Branch.get_list(db=db)
        active_branch_names = {branch.name for branch in active_branches}

        for diff_root in branch_diff_roots:
            if (
                diff_root.base_branch_name != diff_root.diff_branch_name
                and diff_root.diff_branch_name in active_branch_names
                and diff_root.tracking_id
                and isinstance(diff_root.tracking_id, BranchTrackingId)
            ):
                request_diff_update_model = RequestDiffUpdate(branch_name=diff_root.diff_branch_name)
                await get_workflow().submit_workflow(
                    workflow=DIFF_UPDATE, context=context, parameters={"model": request_diff_update_model}
                )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Sequence

from pydantic import BaseModel, ConfigDict, Field
from rich.console import Console
from typing_extensions import Self

from infrahub.core import registry
from infrahub.core.constants import SYSTEM_USER_ID
from infrahub.core.path import SchemaPath  # noqa: TC001
from infrahub.core.query import Query  # noqa: TC001
from infrahub.core.schema import AttributeSchema, MainSchemaTypes, RelationshipSchema, SchemaRoot, internal_schema
from infrahub.core.timestamp import Timestamp

from .query import MigrationBaseQuery  # noqa: TC001

MIGRATION_LOG_TIME_FORMAT = "[%Y-%m-%d %H:%M:%S]"
_migration_console: Console | None = None


def get_migration_console() -> Console:
    global _migration_console

    if _migration_console is None:
        _migration_console = Console(log_time_format=MIGRATION_LOG_TIME_FORMAT)

    return _migration_console


if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase


class MigrationResult(BaseModel):
    errors: list[str] = Field(default_factory=list)
    nbr_migrations_executed: int = 0

    @property
    def success(self) -> bool:
        if not self.errors:
            return True

        return False


@dataclass
class MigrationInput:
    db: InfrahubDatabase
    at: Timestamp = field(default_factory=Timestamp)
    user_id: str = SYSTEM_USER_ID


class SchemaMigration(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(..., description="Name of the migration")
    queries: Sequence[type[MigrationBaseQuery]] = Field(
        ..., description="List of queries to execute for this migration"
    )

    new_node_schema: MainSchemaTypes | None = None
    previous_node_schema: MainSchemaTypes | None = None
    schema_path: SchemaPath

    async def execute_pre_queries(
        self,
        migration_input: MigrationInput,  # noqa: ARG002
        result: MigrationResult,
        branch: Branch,  # noqa: ARG002
    ) -> MigrationResult:
        return result

    async def execute_post_queries(
        self,
        migration_input: MigrationInput,  # noqa: ARG002
        result: MigrationResult,
        branch: Branch,  # noqa: ARG002
    ) -> MigrationResult:
        return result

    async def execute_queries(
        self,
        migration_input: MigrationInput,
        result: MigrationResult,
        branch: Branch,
        queries: Sequence[type[MigrationBaseQuery]],
    ) -> MigrationResult:
        for migration_query in queries:
            try:
                query = await migration_query.init(
                    db=migration_input.db,
                    branch=branch,
                    at=migration_input.at,
                    migration=self,
                    user_id=migration_input.user_id,
                )
                await query.execute(db=migration_input.db)
                result.nbr_migrations_executed += query.get_nbr_migrations_executed()
            except Exception as exc:
                result.errors.append(str(exc))
                return result

        return result

    async def execute(
        self,
        migration_input: MigrationInput,
        branch: Branch,
        queries: Sequence[type[MigrationBaseQuery]] | None = None,
    ) -> MigrationResult:
        async with migration_input.db.start_transaction() as ts:
            result = MigrationResult()
            txn_migration_input = MigrationInput(db=ts, at=migration_input.at, user_id=migration_input.user_id)

            await self.execute_pre_queries(migration_input=txn_migration_input, result=result, branch=branch)
            queries_to_execute = queries or self.queries
            await self.execute_queries(
                migration_input=txn_migration_input, result=result, branch=branch, queries=queries_to_execute
            )
            await self.execute_post_queries(migration_input=txn_migration_input, result=result, branch=branch)

        return result

    @property
    def new_schema(self) -> MainSchemaTypes:
        if self.new_node_schema:
            return self.new_node_schema
        raise ValueError("new_node_schema hasn't been initialized")

    @property
    def previous_schema(self) -> MainSchemaTypes:
        if self.previous_node_schema:
            return self.previous_node_schema
        raise ValueError("previous_node_schema hasn't been initialized")


class AttributeSchemaMigration(SchemaMigration):
    uuids: list[str] | None = None

    @property
    def new_attribute_schema(self) -> AttributeSchema:
        if not self.schema_path.field_name:
            raise ValueError("field_name is not defined")
        return self.new_schema.get_attribute(name=self.schema_path.field_name)

    @property
    def previous_attribute_schema(self) -> AttributeSchema:
        if not self.schema_path.field_name:
            raise ValueError("field_name is not defined")
        return self.previous_schema.get_attribute(name=self.schema_path.field_name)


class RelationshipSchemaMigration(SchemaMigration):
    @property
    def new_relationship_schema(self) -> RelationshipSchema:
        if not self.schema_path.field_name:
            raise ValueError("field_name is not defined")
        return self.new_schema.get_relationship(name=self.schema_path.field_name)

    @property
    def previous_relationship_schema(self) -> RelationshipSchema:
        if not self.schema_path.field_name or not self.previous_node_schema:
            raise ValueError("field_name is not defined")
        return self.previous_schema.get_relationship(name=self.schema_path.field_name)


class GraphMigration(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(..., description="Name of the migration")
    queries: Sequence[type[Query]] = Field(..., description="List of queries to execute for this migration")
    minimum_version: int = Field(..., description="Minimum version of the graph to execute this migration")

    @classmethod
    def init(cls, **kwargs: dict[str, Any]) -> Self:
        return cls(**kwargs)  # type: ignore[arg-type]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        raise NotImplementedError

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        async with migration_input.db.start_transaction() as ts:
            txn_migration_input = MigrationInput(db=ts, at=migration_input.at)
            return await self.do_execute(migration_input=txn_migration_input)

    async def do_execute(self, migration_input: MigrationInput) -> MigrationResult:
        result = MigrationResult()
        for migration_query in self.queries:
            try:
                query = await migration_query.init(db=migration_input.db, at=migration_input.at)
                await query.execute(db=migration_input.db)
            except Exception as exc:
                result.errors.append(str(exc))
                return result

        return result


class InternalSchemaMigration(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(..., description="Name of the migration")
    migrations: Sequence[SchemaMigration] = Field(..., description="")
    minimum_version: int = Field(..., description="Minimum version of the graph to execute this migration")

    @staticmethod
    def get_internal_schema() -> SchemaBranch:
        from infrahub.core.schema.schema_branch import SchemaBranch

        # load the internal schema from
        schema = SchemaRoot(**internal_schema)
        schema_branch = SchemaBranch(cache={}, name="default_branch")
        schema_branch.load_schema(schema=schema)
        schema_branch.process()

        return schema_branch

    @classmethod
    def init(cls, **kwargs: dict[str, Any]) -> Self:
        return cls(**kwargs)  # type: ignore[arg-type]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        raise NotImplementedError

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        result = MigrationResult()

        default_branch = registry.get_branch_from_registry()

        for migration in self.migrations:
            try:
                execution_result = await migration.execute(migration_input=migration_input, branch=default_branch)
                result.errors.extend(execution_result.errors)
            except Exception as exc:
                result.errors.append(str(exc))
                return result

        return result


class ArbitraryMigration(BaseModel):
    name: str = Field(..., description="Name of the migration")
    minimum_version: int = Field(..., description="Minimum version of the graph to execute this migration")

    @classmethod
    def init(cls, **kwargs: dict[str, Any]) -> Self:
        return cls(**kwargs)  # type: ignore[arg-type]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        raise NotImplementedError()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        raise NotImplementedError()


class MigrationRequiringRebase(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(..., description="Name of the migration")
    minimum_version: int = Field(..., description="Minimum version of the graph to execute this migration")

    @classmethod
    def init(cls, **kwargs: dict[str, Any]) -> Self:
        return cls(**kwargs)  # type: ignore[arg-type]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        raise NotImplementedError()

    async def execute_against_branch(self, migration_input: MigrationInput, branch: Branch) -> MigrationResult:
        """Method that will be run against non-default branches, it assumes that the branches have been rebased."""
        raise NotImplementedError()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        """Method that will be run against the default branch."""
        raise NotImplementedError()


type MigrationTypes = GraphMigration | InternalSchemaMigration | ArbitraryMigration | MigrationRequiringRebase



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants import AttributeDBNodeType
from infrahub.core.constants.relationship_label import RELATIONSHIP_TO_NODE_LABEL, RELATIONSHIP_TO_VALUE_LABEL
from infrahub.core.constants.schema import FlagProperty, NodeProperty
from infrahub.core.graph.schema import (
    GraphAttributeIPHostNode,
    GraphAttributeIPNetworkNode,
    GraphAttributeValueIndexedNode,
    GraphAttributeValueNode,
)
from infrahub.core.query import Query, QueryNode, QueryRel, QueryType
from infrahub.core.timestamp import Timestamp
from infrahub.core.utils import build_regex_attrs
from infrahub.types import is_large_attribute_type

if TYPE_CHECKING:
    from infrahub.core.attribute import BaseAttribute
    from infrahub.core.branch import Branch
    from infrahub.core.query import QueryElement
    from infrahub.database import InfrahubDatabase


class AttributeQuery(Query):
    def __init__(
        self,
        attr: BaseAttribute,
        attr_id: str | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | None = None,
        **kwargs: Any,
    ) -> None:
        self.attr = attr
        self.attr_id = attr_id or attr.db_id

        if at:
            self.at = Timestamp(at)
        else:
            self.at = Timestamp()

        self.branch = branch or self.attr.get_branch_based_on_support_type()

        super().__init__(**kwargs)


class AttributeUpdateValueQuery(AttributeQuery):
    name = "attribute_update_value"
    type: QueryType = QueryType.WRITE
    insert_return: bool = False
    raise_error_if_empty: bool = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        at = self.at or self.attr.at

        self.params["attr_uuid"] = self.attr.id
        self.params["user_id"] = self.user_id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = at.to_string()
        content = self.attr.to_db()
        self.params.update(self.attr.to_db())

        prop_list = [f"{key}: ${key}" for key in content.keys()]

        labels = [GraphAttributeValueNode.get_default_label()]
        node_type = self.attr.get_db_node_type()
        if AttributeDBNodeType.INDEXED in node_type:
            labels.append(GraphAttributeValueIndexedNode.get_default_label())
        if AttributeDBNodeType.IPHOST in node_type:
            labels.append(GraphAttributeIPHostNode.get_default_label())
        if AttributeDBNodeType.IPNETWORK in node_type:
            labels.append(GraphAttributeIPNetworkNode.get_default_label())

        query = """
MATCH (a:Attribute { uuid: $attr_uuid })
MERGE (av:%(labels)s { %(props)s } )
WITH av, a
LIMIT 1
// ----------
// find the existing HAS_VALUE edge, if it exists, and set the to time and user_id
// ---------
OPTIONAL MATCH (a)-[existing_active_r:%(rel_label)s { branch: $branch, status: "active" }]->()
WHERE existing_active_r.to IS NULL
SET existing_active_r.to = $at, existing_active_r.to_user_id = $user_id
WITH av, a
LIMIT 1
// ----------
// create the new HAS_VALUE edge
// ---------
CREATE (a)-[r:%(rel_label)s { branch: $branch, branch_level: $branch_level, status: "active", from: $at, from_user_id: $user_id }]->(av)
// ----------
// update the Attribute node with the new timestamp and user id if we are on the default or global branch
// ---------
WITH a
WHERE $branch_level = 1
LIMIT 1
SET a.updated_at = $at, a.updated_by = $user_id
        """ % {"rel_label": self.attr._rel_to_value_label, "labels": ":".join(labels), "props": ", ".join(prop_list)}

        self.add_to_query(query)


class AttributeUpdateFlagQuery(AttributeQuery):
    name = "attribute_update_flag"
    type: QueryType = QueryType.WRITE
    insert_return: bool = False
    raise_error_if_empty: bool = False

    def __init__(
        self,
        flag_name: str,
        **kwargs: Any,
    ) -> None:
        SUPPORTED_FLAGS = ["is_protected"]

        if flag_name not in SUPPORTED_FLAGS:
            raise ValueError(f"Only {SUPPORTED_FLAGS} are supported for now.")

        self.flag_name = flag_name

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        at = self.at or self.attr.at

        self.params["attr_uuid"] = self.attr.id
        self.params["user_id"] = self.user_id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = at.to_string()
        self.params["flag_value"] = getattr(self.attr, self.flag_name)
        self.params["flag_type"] = self.attr.get_kind()

        query = """
MATCH (a:Attribute { uuid: $attr_uuid })
MERGE (flag:Boolean { value: $flag_value })
WITH flag, a
LIMIT 1
// ----------
// find the existing property edge, if it exists, and set the to time and user_id
// ---------
OPTIONAL MATCH (a)-[existing_active_r:%(flag_type)s { branch: $branch, status: "active" }]->()
WHERE existing_active_r.to IS NULL
SET existing_active_r.to = $at, existing_active_r.to_user_id = $user_id
// ----------
// create the new property edge
// ---------
WITH a, flag
CREATE (a)-[r:%(flag_type)s { branch: $branch, branch_level: $branch_level, status: "active", from: $at, from_user_id: $user_id }]->(flag)
// ----------
// update the Attribute node with the new timestamp and user id if we are on the default or global branch
// ---------
WITH a
WHERE $branch_level = 1
LIMIT 1
SET a.updated_at = $at, a.updated_by = $user_id
        """ % {"flag_type": self.flag_name.upper()}
        self.add_to_query(query)


class AttributeUpdateNodePropertyQuery(AttributeQuery):
    name = "attribute_update_node_property"
    type: QueryType = QueryType.WRITE
    insert_return: bool = False
    raise_error_if_empty: bool = False

    def __init__(
        self,
        prop_name: str,
        prop_id: str | None = None,
        **kwargs: Any,
    ) -> None:
        self.prop_name = prop_name
        self.prop_id = prop_id

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        at = self.at or self.attr.at

        branch_filter, branch_params = self.branch.get_query_filter_path(at=at)
        self.params.update(branch_params)
        self.params["attr_uuid"] = self.attr.id
        self.params["user_id"] = self.user_id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = at.to_string()
        self.params["prop_name"] = self.prop_name
        self.params["prop_id"] = self.prop_id

        rel_label = f"HAS_{self.prop_name.upper()}"

        if self.branch.is_default or self.branch.is_global:
            node_query = """
        MATCH (np:Node { uuid: $prop_id })-[r:IS_PART_OF]->(:Root)
        WHERE r.branch IN $branch0
        AND r.status = "active"
        AND r.from <= $at AND (r.to IS NULL OR r.to > $at)
        WITH np
        LIMIT 1
            """
        else:
            node_query = """
        MATCH (np:Node { uuid: $prop_id })-[r:IS_PART_OF]->(:Root)
        WHERE %(branch_filter)s
        ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
        LIMIT 1
        WITH np
        WHERE r.status = "active"
            """ % {"branch_filter": branch_filter}
        self.add_to_query(node_query)

        attr_query = """
MATCH (a:Attribute { uuid: $attr_uuid })
// ----------
// find the existing property edge, if it exists, and set the to time and user_id
// ---------
OPTIONAL MATCH (a)-[existing_active_r:%(rel_label)s { branch: $branch, status: "active" }]->()
WHERE existing_active_r.to IS NULL
SET existing_active_r.to = $at, existing_active_r.to_user_id = $user_id
// ----------
// create the new property edge
// ---------
WITH a, np
LIMIT 1
CREATE (a)-[r:%(rel_label)s { branch: $branch, branch_level: $branch_level, status: "active", from: $at, from_user_id: $user_id }]->(np)
// ----------
// update the Attribute node with the new timestamp and user id if we are on the default or global branch
// ---------
WITH a
WHERE $branch_level = 1
LIMIT 1
SET a.updated_at = $at, a.updated_by = $user_id
        """ % {"rel_label": rel_label}
        self.add_to_query(attr_query)


class AttributeClearNodePropertyQuery(AttributeQuery):
    name = "attribute_clear_node_property"
    type: QueryType = QueryType.WRITE
    insert_return: bool = False

    def __init__(
        self,
        prop_name: str,
        **kwargs: Any,
    ) -> None:
        self.prop_name = prop_name

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        at = self.at or self.attr.at

        branch_filter, branch_params = self.branch.get_query_filter_path(at=at)
        self.params.update(branch_params)
        self.params["attr_uuid"] = self.attr.id
        self.params["user_id"] = self.user_id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["at"] = at.to_string()

        rel_label = f"HAS_{self.prop_name.upper()}"
        query = """
MATCH (a:Attribute { uuid: $attr_uuid })-[r:%(rel_label)s]->(np:Node)
WITH DISTINCT a, np
CALL (a, np) {
    MATCH (a)-[r:%(rel_label)s]->(np)
    WHERE %(branch_filter)s
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    RETURN r AS property_edge
}
WITH a, np, property_edge
WHERE property_edge.status = "active"
CALL (property_edge) {
    WITH property_edge
    WHERE property_edge.branch = $branch
    SET property_edge.to = $at, property_edge.to_user_id = $user_id
}
CALL (a, np, property_edge) {
    WITH property_edge
    WHERE property_edge.branch_level < $branch_level
    CREATE (a)-[r:%(rel_label)s { branch: $branch, branch_level: $branch_level, status: "deleted", from: $at, from_user_id: $user_id }]->(np)
}
CALL (a) {
    WITH a
    WHERE $branch_level = 1
    LIMIT 1
    SET a.updated_at = $at, a.updated_by = $user_id
}
        """ % {"branch_filter": branch_filter, "rel_label": rel_label}
        self.add_to_query(query)


class AttributeDeleteQuery(AttributeQuery):
    name = "attribute_delete"
    type: QueryType = QueryType.WRITE
    insert_return: bool = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params["attr_uuid"] = self.attr.id
        self.params["user_id"] = self.user_id
        self.params["branch"] = self.branch.name
        self.params["branch_level"] = self.branch.hierarchy_level
        self.params["branched_from"] = self.branch.get_branched_from()
        self.params["at"] = self.at.to_string()

        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at)
        self.params.update(branch_params)

        query = """
MATCH (a:Attribute { uuid: $attr_uuid })
CALL (a) {
    WITH a
    WHERE $branch_level = 1
    LIMIT 1
    SET a.updated_at = $at, a.updated_by = $user_id
}

UNWIND [
    ["HAS_ATTRIBUTE", "in"],
    ["HAS_VALUE", "out"],
    ["IS_PROTECTED", "out"],
    ["HAS_SOURCE", "out"],
    ["HAS_OWNER", "out"]
] AS edge_details
WITH a, edge_details[0] AS property_type, edge_details[1] AS direction
CALL (a, property_type, direction) {
    MATCH (a)-[r]-(attr_peer)
    WHERE type(r) = property_type
    AND (
        (direction = "in" AND startNode(r) = attr_peer)
        OR (direction = "out" AND startNode(r) = a)
    )
    AND %(branch_filter)s
    RETURN r AS property_edge, attr_peer
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
CALL (property_edge) {
    WITH property_edge
    WHERE property_edge.status = "active"
    AND property_edge.branch = $branch
    AND property_edge.to IS NULL
    SET property_edge.to = $at, property_edge.to_user_id = $user_id
}
WITH a, property_edge, property_type, attr_peer, direction
CALL (a, property_type, attr_peer, direction) {
    WITH direction
    WHERE direction = "out"
    CREATE (a)
        -[r:$(property_type) { branch: $branch, branch_level: $branch_level, status: "deleted", from: $at, from_user_id: $user_id }]
        ->(attr_peer)
}
CALL (a, property_type, attr_peer, direction) {
    WITH direction
    WHERE direction = "in"
    CREATE (a)
        <-[r:$(property_type) { branch: $branch, branch_level: $branch_level, status: "deleted", from: $at, from_user_id: $user_id }]
        -(attr_peer)
}
WITH CASE
    WHEN property_type = "HAS_VALUE" THEN attr_peer.value
    ELSE NULL
END AS property_value
WITH property_value
RETURN property_value
ORDER BY property_value ASC
LIMIT 1
        """ % {"branch_filter": branch_filter}
        self.add_to_query(query)
        self.return_labels = ["property_value"]

    def get_previous_property_value(self) -> Any:
        result = self.get_result()
        if result:
            return result.get(label="property_value")
        return None


async def default_attribute_query_filter(
    name: str,
    filter_name: str,
    branch: Branch | None = None,  # noqa: ARG001
    filter_value: str | int | bool | list | None = None,
    attribute_kind: str | None = None,
    include_match: bool = True,
    param_prefix: str | None = None,
    db: InfrahubDatabase | None = None,  # noqa: ARG001
    partial_match: bool = False,
) -> tuple[list[QueryElement], dict[str, Any], list[str]]:
    """Generate Query String Snippet to filter the right node."""
    attribute_value_label = GraphAttributeValueNode.get_default_label()
    if attribute_kind and not is_large_attribute_type(attribute_kind):
        attribute_value_label = GraphAttributeValueIndexedNode.get_default_label()

    query_filter: list[QueryElement] = []
    query_params: dict[str, Any] = {}
    query_where: list[str] = []

    if filter_value and not isinstance(filter_value, str | bool | int | list):
        raise TypeError(f"filter {filter_name}: {filter_value} ({type(filter_value)}) is not supported.")

    if isinstance(filter_value, list) and not all(isinstance(value, str | bool | int) for value in filter_value):
        raise TypeError(f"filter {filter_name}: {filter_value} (list) contains unsupported item")

    param_prefix = param_prefix or f"attr_{name}"

    if include_match:
        query_filter.append(QueryNode(name="n"))

    query_filter.append(QueryRel(labels=[RELATIONSHIP_TO_NODE_LABEL]))

    if name in ["any", "attribute"]:
        query_filter.append(QueryNode(name="i", labels=["Attribute"]))
    else:
        query_filter.append(QueryNode(name="i", labels=["Attribute"], params={"name": f"${param_prefix}_name"}))
        query_params[f"{param_prefix}_name"] = name

    if filter_name in ("value", "binary_address", "prefixlen", "isnull"):
        query_filter.append(QueryRel(labels=[RELATIONSHIP_TO_VALUE_LABEL]))

        if filter_value is None:
            query_filter.append(QueryNode(name="av", labels=[attribute_value_label]))
        else:
            if partial_match:
                query_filter.append(QueryNode(name="av", labels=[attribute_value_label]))
                query_where.append(
                    f"toLower(toString(av.{filter_name})) CONTAINS toLower(toString(${param_prefix}_{filter_name}))"
                )
            elif attribute_kind and attribute_kind == "List" and not isinstance(filter_value, list):
                query_filter.append(QueryNode(name="av", labels=[attribute_value_label]))
                filter_value = build_regex_attrs(values=[filter_value])
                query_where.append(f"toString(av.{filter_name}) =~ ${param_prefix}_{filter_name}")
            elif filter_name == "isnull":
                query_filter.append(QueryNode(name="av", labels=[attribute_value_label]))
            else:
                query_filter.append(
                    QueryNode(
                        name="av",
                        labels=[attribute_value_label],
                        params={filter_name: f"${param_prefix}_{filter_name}"},
                    )
                )
            query_params[f"{param_prefix}_{filter_name}"] = filter_value

    elif filter_name == "values" and isinstance(filter_value, list):
        query_filter.extend(
            (QueryRel(labels=[RELATIONSHIP_TO_VALUE_LABEL]), QueryNode(name="av", labels=[attribute_value_label]))
        )
        if attribute_kind and attribute_kind == "List":
            query_params[f"{param_prefix}_{filter_name}"] = build_regex_attrs(values=filter_value)
            query_where.append(f"toString(av.value) =~ ${param_prefix}_{filter_name}")
        else:
            query_where.append(f"av.value IN ${param_prefix}_value")
        query_params[f"{param_prefix}_value"] = filter_value

    elif filter_name == "version":
        query_filter.append(QueryRel(labels=[RELATIONSHIP_TO_VALUE_LABEL]))

        if filter_value is None:
            query_filter.append(QueryNode(name="av", labels=[GraphAttributeValueNode.get_default_label()]))
        else:
            query_filter.append(
                QueryNode(
                    name="av",
                    labels=[GraphAttributeValueNode.get_default_label()],
                    params={filter_name: f"${param_prefix}_{filter_name}"},
                )
            )
            query_params[f"{param_prefix}_{filter_name}"] = filter_value

    elif filter_name in [v.value for v in FlagProperty] and filter_value is not None:
        query_filter.append(QueryRel(labels=[filter_name.upper()]))
        query_filter.append(
            QueryNode(name="ap", labels=["Boolean"], params={"value": f"${param_prefix}_{filter_name}"})
        )
        query_params[f"{param_prefix}_{filter_name}"] = filter_value

    elif "__" in filter_name and filter_value is not None:
        filter_name_split = filter_name.split(sep="__", maxsplit=1)
        property_name: str = filter_name_split[0]
        property_attr: str = filter_name_split[1]

        if property_name not in [v.value for v in NodeProperty]:
            raise ValueError(f"filter {filter_name}: {filter_value}, {property_name} is not a valid property")

        if property_attr != "id":
            raise ValueError(f"filter {filter_name}: {filter_value}, {property_attr} is supported")

        clean_filter_name = f"{property_name}_{property_attr}"

        query_filter.extend(
            [
                QueryRel(labels=[f"HAS_{property_name.upper()}"]),
                QueryNode(name="ap", labels=["Node"], params={"uuid": f"${param_prefix}_{clean_filter_name}"}),
            ]
        )
        query_params[f"{param_prefix}_{clean_filter_name}"] = filter_value

    return query_filter, query_params, query_where



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub_sdk.template import Jinja2Template
from rich.progress import Progress, TaskID

from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME, NULL_VALUE, BranchSupportType
from infrahub.core.initialization import get_root_node
from infrahub.core.migrations.shared import (
    MigrationInput,
    MigrationRequiringRebase,
    MigrationResult,
    get_migration_console,
)
from infrahub.core.query import Query, QueryType

from .load_schema_branch import get_or_load_schema_branch
from .m044_backfill_hfid_display_label_in_db import (
    DefaultBranchNodeCount,
    GetPathDetailsBranchQuery,
    GetPathDetailsDefaultBranch,
    GetResultMapQuery,
)

if TYPE_CHECKING:
    from infrahub.core.schema import AttributeSchema, MainSchemaTypes
    from infrahub.core.schema.basenode_schema import SchemaAttributePath
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


console = get_migration_console()


def _is_jinja2_template(display_label: str) -> bool:
    return any(c in display_label for c in "{}")


def _extract_jinja2_variables(template_str: str) -> list[str]:
    return Jinja2Template(template=template_str).get_variables()


async def _render_display_label(display_label: str, variable_names: list[str], values: list[Any]) -> str | None:
    if not _is_jinja2_template(display_label):
        return values[0] if values and values[0] is not None else None

    variables = dict(zip(variable_names, values, strict=False))
    jinja_template = Jinja2Template(template=display_label)
    return await jinja_template.render(variables=variables)


class UpdateAttributeValuesQuery(Query):
    """
    Update the values of the given attribute schema for the input node-id-to-value map.

    This version only expires existing values when they're different from the new value,
    making it safe to run idempotently without clearing correct existing values.

    This code is adapted from m044_backfill_hfid_display_label_in_db.
    """

    name = "update_attribute_values"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, attribute_schema: AttributeSchema, values_by_id_map: dict[str, Any], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.attribute_name = attribute_schema.name
        self.is_branch_agnostic = attribute_schema.get_branch() is BranchSupportType.AGNOSTIC
        self.values_by_id_map = values_by_id_map

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params = {
            "node_uuids": list(self.values_by_id_map.keys()),
            "attribute_name": self.attribute_name,
            "values_by_id": self.values_by_id_map,
            "default_branch": registry.default_branch,
            "global_branch": GLOBAL_BRANCH_NAME,
            "branch": GLOBAL_BRANCH_NAME if self.is_branch_agnostic else self.branch.name,
            "branch_level": 1 if self.is_branch_agnostic else self.branch.hierarchy_level,
            "at": self.at.to_string(),
        }
        branch_filter, branch_filter_params = self.branch.get_query_filter_path(at=self.at)
        self.params.update(branch_filter_params)

        if self.branch.name in [registry.default_branch, GLOBAL_BRANCH_NAME]:
            update_value_query = """
// ------------
// Find the Nodes and Attributes we need to update
// ------------
MATCH (n:Node)-[e:IS_PART_OF]->(:Root)
WHERE n.uuid IN $node_uuids
AND e.branch IN [$default_branch, $global_branch]
AND e.to IS NULL
AND e.status = "active"
WITH DISTINCT n
MATCH (n)-[e:HAS_ATTRIBUTE]->(attr:Attribute {name: $attribute_name})
WHERE e.branch IN [$default_branch, $global_branch]
AND e.to IS NULL
AND e.status = "active"
// ------------
// If the attribute has an existing value on the branch, then set the to time on it
// but only if the value is different from the new value
// ------------
WITH DISTINCT n, attr
CALL (attr) {
    OPTIONAL MATCH (attr)-[e:HAS_VALUE]->(existing_av)
    WHERE e.branch IN [$default_branch, $global_branch]
    AND e.to IS NULL
    AND e.status = "active"
    RETURN existing_av, e AS existing_has_value
}
CALL (existing_has_value, existing_av, n) {
    WITH existing_has_value, existing_av, n
    WHERE existing_has_value IS NOT NULL
    AND existing_av.value <> $values_by_id[n.uuid]
    SET existing_has_value.to = $at
}
WITH n, attr, existing_av
            """
        else:
            update_value_query = """
// ------------
// Find the Nodes and Attributes we need to update
// ------------
MATCH (n:Node)
WHERE n.uuid IN $node_uuids
CALL (n) {
    MATCH (n)-[r:IS_PART_OF]->(:Root)
    WHERE %(branch_filter)s
    RETURN r.status = "active" AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH n, is_active
WHERE is_active = TRUE
WITH DISTINCT n
CALL (n) {
    MATCH (n)-[r:HAS_ATTRIBUTE]->(attr:Attribute {name: $attribute_name})
    WHERE %(branch_filter)s
    RETURN attr, r.status = "active"  AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH DISTINCT n, attr, is_active
WHERE is_active = TRUE
// ------------
// If the attribute has an existing value on the branch, then set the to time on it
// but only if the value is different from the new value
// ------------
CALL (n, attr) {
    OPTIONAL MATCH (attr)-[r:HAS_VALUE]->(existing_av)
    WHERE %(branch_filter)s
    WITH r, existing_av, n
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    WITH CASE
        WHEN existing_av.value <> $values_by_id[n.uuid]
        AND r.status = "active"
        AND r.branch = $branch
        THEN [r, existing_av]
        ELSE [NULL, NULL]
    END AS existing_details
    RETURN existing_details[0] AS existing_has_value, existing_details[1] AS existing_av
}
CALL (existing_has_value) {
    WITH existing_has_value
    WHERE existing_has_value IS NOT NULL
    SET existing_has_value.to = $at
}
WITH n, attr, existing_av
            """ % {"branch_filter": branch_filter}
        self.add_to_query(update_value_query)

        set_value_query = """
// ------------
// only make updates if the existing value is not the same as the new value
// ------------
WITH n, attr, existing_av, $values_by_id[n.uuid] AS required_value
WHERE existing_av.value <> required_value
OR existing_av IS NULL
CALL (n, attr) {
    MERGE (av:AttributeValue&AttributeValueIndexed {is_default: false, value: $values_by_id[n.uuid]} )
    WITH av, attr
    LIMIT 1
    CREATE (attr)-[r:HAS_VALUE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(av)
}
            """
        self.add_to_query(set_value_query)


class GetNodesWithoutDisplayLabelQuery(Query):
    """Get all active nodes that do not have a display_label attribute on the default branch."""

    name = "get_nodes_without_display_label"
    type = QueryType.READ

    def __init__(self, kinds_to_skip: list[str] | None = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.kinds_to_skip = kinds_to_skip or []

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params = {
            "branch_names": [registry.default_branch, GLOBAL_BRANCH_NAME],
            "kinds_to_skip": self.kinds_to_skip,
            "attribute_name": "display_label",
        }
        query = """
// ------------
// Get all active nodes that don't have a display_label attribute
// ------------
MATCH (n:Node)-[e:IS_PART_OF]->(:Root)
WHERE NOT n.kind IN $kinds_to_skip
AND e.branch IN $branch_names
AND e.status = "active"
AND e.to IS NULL
AND NOT exists((n)-[:IS_PART_OF {branch: e.branch, status: "deleted"}]->(:Root))
WITH DISTINCT n
CALL (n) {
    OPTIONAL MATCH (n)-[r:HAS_ATTRIBUTE]->(attr:Attribute {name: $attribute_name})
    WHERE r.branch IN $branch_names
    RETURN r AS has_attr_e
    ORDER BY r.from DESC, r.status ASC
    LIMIT 1
}
WITH n, has_attr_e
WHERE (has_attr_e IS NULL OR has_attr_e.status = "deleted")
WITH n.uuid AS node_uuid
        """
        self.add_to_query(query)
        self.return_labels = ["node_uuid"]

    def get_node_uuids(self) -> list[str]:
        return [result.get_as_type(label="node_uuid", return_type=str) for result in self.get_results()]


class GetNodesWithoutDisplayLabelBranchQuery(Query):
    """Get all active nodes that do not have a display_label attribute on a non-default branch."""

    name = "get_nodes_without_display_label_branch"
    type = QueryType.READ

    def __init__(self, kinds_to_skip: list[str] | None = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.kinds_to_skip = kinds_to_skip or []

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_filter_params = self.branch.get_query_filter_path(at=self.at)
        self.params = {
            "kinds_to_skip": self.kinds_to_skip,
            "attribute_name": "display_label",
            **branch_filter_params,
        }
        query = """
// ------------
// Get all active nodes that don't have a display_label attribute
// ------------
MATCH (n:Node)
WHERE NOT n.kind IN $kinds_to_skip
CALL (n) {
    MATCH (n)-[r:IS_PART_OF]->(:Root)
    WHERE %(branch_filter)s
    RETURN r AS is_part_of_e
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH n, is_part_of_e
WHERE is_part_of_e.status = "active"
CALL (n) {
    OPTIONAL MATCH (n)-[r:HAS_ATTRIBUTE]->(attr:Attribute {name: $attribute_name})
    WHERE %(branch_filter)s
    RETURN r AS has_attr_e
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH n, has_attr_e
WHERE (has_attr_e IS NULL OR has_attr_e.status = "deleted")
WITH n.uuid AS node_uuid
        """ % {"branch_filter": branch_filter}
        self.add_to_query(query)
        self.return_labels = ["node_uuid"]

    def get_node_uuids(self) -> list[str]:
        return [result.get_as_type(label="node_uuid", return_type=str) for result in self.get_results()]


class CreateDisplayLabelNullQuery(Query):
    """Create display_label attribute with NULL value for the given nodes."""

    name = "create_display_label_null"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, node_uuids: list[str], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.node_uuids = node_uuids

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params = {
            "node_uuids": self.node_uuids,
            "attribute_name": "display_label",
            "default_branch": registry.default_branch,
            "global_branch": GLOBAL_BRANCH_NAME,
            "branch": self.branch.name,
            "branch_level": self.branch.hierarchy_level,
            "at": self.at.to_string(),
            "null_value": NULL_VALUE,
            "branch_support": "aware",
            "is_protected_default": False,
            "is_visible_default": True,
        }
        branch_filter, branch_filter_params = self.branch.get_query_filter_path(at=self.at)
        self.params.update(branch_filter_params)

        # Create the NULL AttributeValue first
        create_av_query = """
MERGE (av:AttributeValue&AttributeValueIndexed {is_default: true, value: $null_value})
WITH av
LIMIT 1
MERGE (is_protected_value:Boolean { value: $is_protected_default })
MERGE (is_visible_value:Boolean { value: $is_visible_default })
        """
        self.add_to_query(create_av_query)

        if self.branch.name in [registry.default_branch, GLOBAL_BRANCH_NAME]:
            query = """
// ------------
// Create the display_label attribute with NULL value for nodes
// ------------
WITH av, is_protected_value, is_visible_value
MATCH (n:Node)-[e:IS_PART_OF]->(:Root)
WHERE n.uuid IN $node_uuids
AND e.branch IN [$default_branch, $global_branch]
AND e.to IS NULL
AND e.status = "active"
CREATE (a:Attribute { uuid: randomUUID(), name: $attribute_name, branch_support: $branch_support })
CREATE (n)-[:HAS_ATTRIBUTE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(a)
CREATE (a)-[:HAS_VALUE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(av)
CREATE (a)-[:IS_PROTECTED { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(is_protected_value)
CREATE (a)-[:IS_VISIBLE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(is_visible_value)
            """
        else:
            query = """
// ------------
// Create the display_label attribute with NULL value for nodes
// ------------
WITH av, is_protected_value, is_visible_value
MATCH (n:Node)
WHERE n.uuid IN $node_uuids
CALL (n) {
    MATCH (n)-[r:IS_PART_OF]->(:Root)
    WHERE %(branch_filter)s
    RETURN r.status = "active" AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH n, is_active, av, is_protected_value, is_visible_value
WHERE is_active = TRUE
CREATE (a:Attribute { uuid: randomUUID(), name: $attribute_name, branch_support: $branch_support })
CREATE (n)-[:HAS_ATTRIBUTE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(a)
CREATE (a)-[:HAS_VALUE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(av)
CREATE (a)-[:IS_PROTECTED { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(is_protected_value)
CREATE (a)-[:IS_VISIBLE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(is_visible_value)
            """ % {"branch_filter": branch_filter}

        self.add_to_query(query)


class Migration047(MigrationRequiringRebase):
    """
    Backfill `display_label` attributes for all nodes:
    - If schema does not define display_label OR attribute doesn't exist: insert NULL value
    - If schema defines display_label: compute and store the value, invalidate NULL value if exists
    """

    name: str = "047_backfill_or_null_display_label"
    minimum_version: int = 46
    update_batch_size: int = 1000
    # skip these b/c the attributes on these schema-related nodes are used to define the values included in
    # the display_label attributes on instances of these schema, so should not be updated
    kinds_to_skip: list[str] = ["SchemaNode", "SchemaAttribute", "SchemaRelationship", "SchemaGeneric"]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    def _extract_schema_paths_from_display_label(
        self, schema: MainSchemaTypes, schema_branch: SchemaBranch
    ) -> list[SchemaAttributePath]:
        """Extract schema paths from display_label, handling both simple paths and Jinja2 templates.

        This follows the same logic as _validate_display_label in schema_branch.py.
        """
        if not schema.display_label:
            return []

        if not _is_jinja2_template(schema.display_label):
            schema_path = schema.parse_schema_path(path=schema.display_label, schema=schema_branch)
            return [schema_path]

        schema_paths = []
        for variable in _extract_jinja2_variables(schema.display_label):
            schema_path = schema.parse_schema_path(path=variable, schema=schema_branch)
            schema_paths.append(schema_path)

        return schema_paths

    async def _do_one_schema_all(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        schema: MainSchemaTypes,
        schema_branch: SchemaBranch,
        attribute_schema: AttributeSchema,
        at: Timestamp,
        progress: Progress | None = None,
        update_task: TaskID | None = None,
    ) -> None:
        if not schema.display_label:
            return

        schema_paths = self._extract_schema_paths_from_display_label(schema=schema, schema_branch=schema_branch)
        if not schema_paths:
            return

        offset = 0

        # loop until we get no results from the get_details_query
        while True:
            if branch.is_default:
                get_details_query: GetResultMapQuery = await GetPathDetailsDefaultBranch.init(
                    db=db,
                    schema_kind=schema.kind,
                    schema_paths=schema_paths,
                    offset=offset,
                    limit=self.update_batch_size,
                )
            else:
                get_details_query = await GetPathDetailsBranchQuery.init(
                    db=db,
                    branch=branch,
                    schema_kind=schema.kind,
                    schema_paths=schema_paths,
                    updates_only=False,
                    offset=offset,
                    limit=self.update_batch_size,
                )
            await get_details_query.execute(db=db)

            # Get the values for all schema paths
            schema_path_values_map = get_details_query.get_result_map(schema_paths)
            num_updates = len(schema_path_values_map)

            formatted_schema_path_values_map: dict[str, str] = {}
            for k, v in schema_path_values_map.items():
                if not v:
                    continue

                rendered_value = await _render_display_label(
                    display_label=schema.display_label,
                    variable_names=[s.attribute_path_as_str for s in schema_paths],
                    values=v,
                )
                if rendered_value is not None:
                    formatted_schema_path_values_map[k] = rendered_value

            if formatted_schema_path_values_map:
                update_display_label_query = await UpdateAttributeValuesQuery.init(
                    db=db,
                    branch=branch,
                    attribute_schema=attribute_schema,
                    values_by_id_map=formatted_schema_path_values_map,
                    at=at,
                )
                await update_display_label_query.execute(db=db)

            if progress is not None and update_task is not None:
                progress.update(update_task, advance=num_updates)

            if num_updates == 0:
                break

            offset += self.update_batch_size

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        at = migration_input.at
        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)

        main_schema_branch = await get_or_load_schema_branch(db=db, branch=default_branch)

        base_node_schema = main_schema_branch.get("SchemaNode", duplicate=False)
        display_label_attribute_schema = base_node_schema.get_attribute("display_label")

        # Get nodes without display_label in the database
        get_nodes_without_dl_query = await GetNodesWithoutDisplayLabelQuery.init(
            db=db, kinds_to_skip=self.kinds_to_skip
        )
        await get_nodes_without_dl_query.execute(db=db)
        nodes_without_display_label = get_nodes_without_dl_query.get_node_uuids()

        # Count nodes that will get computed values
        kinds_to_backfill: list[str] = []
        for node_schema_name in (
            main_schema_branch.node_names + main_schema_branch.profile_names + main_schema_branch.template_names
        ):
            if node_schema_name in self.kinds_to_skip:
                continue

            node_schema = main_schema_branch.get(name=node_schema_name, duplicate=False)
            if node_schema.branch != BranchSupportType.AWARE or not node_schema.display_label:
                continue

            kinds_to_backfill.append(node_schema.kind)

        backfill_count = 0
        if kinds_to_backfill:
            count_query = await DefaultBranchNodeCount.init(
                db=db, kinds_to_skip=self.kinds_to_skip, kinds_to_include=kinds_to_backfill
            )
            await count_query.execute(db=db)
            backfill_count = count_query.get_num_nodes()

        try:
            with Progress(console=console) as progress:
                # Create NULL display_label
                if nodes_without_display_label:
                    null_task = progress.add_task(
                        f"Creating NULL display_label for {len(nodes_without_display_label)} nodes",
                        total=len(nodes_without_display_label),
                    )

                    for offset in range(0, len(nodes_without_display_label), self.update_batch_size):
                        batch_uuids = nodes_without_display_label[offset : offset + self.update_batch_size]
                        if not batch_uuids:
                            break

                        create_display_label_query = await CreateDisplayLabelNullQuery.init(
                            db=db, branch=default_branch, node_uuids=batch_uuids, at=at
                        )
                        await create_display_label_query.execute(db=db)

                        progress.update(null_task, advance=len(batch_uuids))

                # Backfill computed display_label values
                if backfill_count > 0:
                    backfill_task = progress.add_task(
                        f"Backfilling computed display_label for {backfill_count} nodes",
                        total=backfill_count,
                    )

                    for node_schema_name in kinds_to_backfill:
                        await self._do_one_schema_all(
                            db=db,
                            branch=default_branch,
                            schema=main_schema_branch.get(name=node_schema_name, duplicate=False),
                            schema_branch=main_schema_branch,
                            attribute_schema=display_label_attribute_schema,
                            at=at,
                            progress=progress,
                            update_task=backfill_task,
                        )

        except Exception as exc:
            return MigrationResult(errors=[str(exc)])
        return MigrationResult()

    async def execute_against_branch(self, migration_input: MigrationInput, branch: Branch) -> MigrationResult:
        db = migration_input.db
        at = migration_input.at
        schema_branch = await get_or_load_schema_branch(db=db, branch=branch)

        base_node_schema = schema_branch.get("SchemaNode", duplicate=False)
        display_label_attribute_schema = base_node_schema.get_attribute("display_label")

        try:
            get_nodes_without_dl_query = await GetNodesWithoutDisplayLabelBranchQuery.init(
                db=db, branch=branch, kinds_to_skip=self.kinds_to_skip
            )
            await get_nodes_without_dl_query.execute(db=db)
            nodes_without_display_label = get_nodes_without_dl_query.get_node_uuids()

            if nodes_without_display_label:
                for offset in range(0, len(nodes_without_display_label), self.update_batch_size):
                    batch_uuids = nodes_without_display_label[offset : offset + self.update_batch_size]
                    if not batch_uuids:
                        break

                    create_display_label_query = await CreateDisplayLabelNullQuery.init(
                        db=db, branch=branch, node_uuids=batch_uuids, at=at
                    )
                    await create_display_label_query.execute(db=db)

            for node_schema_name in (
                schema_branch.node_names + schema_branch.profile_names + schema_branch.template_names
            ):
                if node_schema_name in self.kinds_to_skip:
                    continue

                node_schema = schema_branch.get(name=node_schema_name, duplicate=False)
                if node_schema.branch != BranchSupportType.AWARE or not node_schema.display_label:
                    continue

                await self._do_one_schema_all(
                    db=db,
                    branch=branch,
                    schema=node_schema,
                    schema_branch=schema_branch,
                    attribute_schema=display_label_attribute_schema,
                    at=at,
                )

        except Exception as exc:
            return MigrationResult(errors=[str(exc)])
        return MigrationResult()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from infrahub.core.constants import NULL_VALUE, PathType
from infrahub.core.path import DataPath, GroupedDataPaths
from infrahub.exceptions import ValidationError
from infrahub.types import get_attribute_type

from ..interface import ConstraintCheckerInterface
from ..shared import AttributeSchemaValidatorQuery

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase

    from ..model import SchemaConstraintValidatorRequest


@dataclass
class NodeAttributeValue:
    node_id: str
    attribute_value: Any
    branch_name: str


class AttributeKindUpdateValidatorQuery(AttributeSchemaValidatorQuery):
    name: str = "attribute_constraints_kind_validator"

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)

        self.params["attr_name"] = self.attribute_schema.name
        self.params["null_value"] = NULL_VALUE

        query = """
        MATCH (n:%(node_kinds)s)
        CALL (n) {
            MATCH path = (root:Root)<-[rr:IS_PART_OF]-(n)-[ra:HAS_ATTRIBUTE]-(:Attribute { name: $attr_name } )-[rv:HAS_VALUE]-(av:AttributeValue)
            WHERE all(
                r in relationships(path)
                WHERE %(branch_filter)s
            )
            RETURN path as full_path, n as node, rv as value_relationship, av.value as attribute_value
            ORDER BY rv.branch_level DESC, ra.branch_level DESC, rr.branch_level DESC, rv.from DESC, ra.from DESC, rr.from DESC
            LIMIT 1
        }
        WITH full_path, node, attribute_value, value_relationship
        WHERE all(r in relationships(full_path) WHERE r.status = "active")
        AND attribute_value IS NOT NULL
        AND attribute_value <> $null_value
        """ % {
            "branch_filter": branch_filter,
            "node_kinds": f"{self.node_schema.kind}|Profile{self.node_schema.kind}|Template{self.node_schema.kind}",
        }

        self.add_to_query(query)
        self.return_labels = ["node.uuid", "attribute_value", "value_relationship.branch as value_branch"]

    async def get_paths(self) -> GroupedDataPaths:
        grouped_data_paths = GroupedDataPaths()
        infrahub_data_type = get_attribute_type(self.attribute_schema.kind)
        infrahub_attribute_class = infrahub_data_type.get_infrahub_class()
        for result in self.get_results():
            value = result.get("attribute_value")
            if value in (None, NULL_VALUE):
                continue
            try:
                attr_value = result.get("attribute_value")
                infrahub_attribute_class.validate_format(
                    value=attr_value, name=self.attribute_schema.name, schema=self.attribute_schema
                )
                infrahub_attribute_class.validate_content(
                    value=attr_value, name=self.attribute_schema.name, schema=self.attribute_schema
                )
            except ValidationError:
                grouped_data_paths.add_data_path(
                    DataPath(
                        branch=str(result.get("value_branch")),
                        path_type=PathType.ATTRIBUTE,
                        node_id=str(result.get("node.uuid")),
                        field_name=self.attribute_schema.name,
                        kind=self.node_schema.kind,
                        value=value,
                    )
                )
        return grouped_data_paths


class AttributeKindChecker(ConstraintCheckerInterface):
    query_classes = [AttributeKindUpdateValidatorQuery]

    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None) -> None:
        self.db = db
        self.branch = branch

    @property
    def name(self) -> str:
        return "attribute.kind.update"

    def supports(self, request: SchemaConstraintValidatorRequest) -> bool:
        return request.constraint_name == self.name

    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]:
        grouped_data_paths_list: list[GroupedDataPaths] = []
        for query_class in self.query_classes:
            # TODO add exception handling
            query = await query_class.init(
                db=self.db, branch=self.branch, node_schema=request.node_schema, schema_path=request.schema_path
            )

            await query.execute(db=self.db)
            grouped_data_paths_list.append(await query.get_paths())
        return grouped_data_paths_list



from __future__ import annotations

import ipaddress
from dataclasses import dataclass
from typing import TYPE_CHECKING

from infrahub.core.constants import InfrahubKind
from infrahub.core.graph.schema import GraphAttributeIPHostNode, GraphAttributeIPNetworkNode
from infrahub.core.ipam.constants import AllIPTypes, IPAddressType, IPNetworkType
from infrahub.core.query import Query, QueryResult, QueryType
from infrahub.core.registry import registry
from infrahub.core.utils import convert_ip_to_binary_str

if TYPE_CHECKING:
    from uuid import UUID

    from infrahub.core.node import Node
    from infrahub.database import InfrahubDatabase


PREFIX_ATTRIBUTE_LABEL = GraphAttributeIPNetworkNode.get_default_label()
ADDRESS_ATTRIBUTE_LABEL = GraphAttributeIPHostNode.get_default_label()


@dataclass
class IPPrefixData:
    id: UUID
    prefix: IPNetworkType


@dataclass
class IPAddressData:
    id: UUID
    address: IPAddressType


@dataclass(frozen=True)
class IPPrefixFreeData:
    free_start: int

    @classmethod
    def from_db(cls, result: QueryResult) -> IPPrefixFreeData:
        return cls(
            free_start=result.get_as_type("free_start", return_type=int),
        )


@dataclass(frozen=True)
class IPv6PrefixFreeData:
    free_start_bin: str

    @classmethod
    def from_db(cls, result: QueryResult) -> IPv6PrefixFreeData:
        return cls(
            free_start_bin=result.get_as_type("free_start_bin", return_type=str),
        )


@dataclass(frozen=True)
class IPAddressFreeData:
    free_addr: int
    is_free: bool
    is_last: bool

    @classmethod
    def from_db(cls, result: QueryResult) -> IPAddressFreeData:
        return cls(
            free_addr=result.get_as_type("free_addr", return_type=int),
            is_free=result.get_as_type("is_free", return_type=bool),
            is_last=result.get_as_type("is_last", return_type=bool),
        )


@dataclass(frozen=True)
class IPv6AddressFreeData:
    free_addr_bin: str
    is_free: bool
    is_last: bool

    @classmethod
    def from_db(cls, result: QueryResult) -> IPv6AddressFreeData:
        return cls(
            free_addr_bin=result.get_as_type("free_addr_bin", return_type=str),
            is_free=result.get_as_type("is_free", return_type=bool),
            is_last=result.get_as_type("is_last", return_type=bool),
        )


def _get_namespace_id(
    namespace: Node | str | None = None,
) -> str:
    if namespace and isinstance(namespace, str):
        return namespace
    if namespace and hasattr(namespace, "id"):
        return namespace.id
    return registry.default_ipnamespace


class IPPrefixSubnetFetch(Query):
    name = "ipprefix_subnet_fetch"
    type = QueryType.READ

    def __init__(
        self,
        obj: IPNetworkType,
        namespace: Node | str | None = None,
        **kwargs,
    ) -> None:
        self.obj = obj
        self.namespace_id = _get_namespace_id(namespace)

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["ns_id"] = self.namespace_id

        prefix_bin = convert_ip_to_binary_str(self.obj)[: self.obj.prefixlen]
        self.params["prefix_binary"] = prefix_bin
        self.params["maxprefixlen"] = self.obj.prefixlen
        self.params["ip_version"] = self.obj.version

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)

        # ruff: noqa: E501
        query = """
        // First match on IPNAMESPACE
        MATCH (ns:%(ns_label)s)
        WHERE ns.uuid = $ns_id
        CALL (ns) {
            MATCH (ns)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN r
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH ns, r
        WHERE r.status = "active"
        WITH ns
        // MATCH all prefixes that are IN SCOPE
        MATCH path2 = (ns)-[:IS_RELATED]-(ns_rel:Relationship)-[:IS_RELATED]-(pfx:%(node_label)s)-[:HAS_ATTRIBUTE]-(an:Attribute {name: "prefix"})-[:HAS_VALUE]-(av:AttributeIPNetwork)
        WHERE ns_rel.name = "ip_namespace__ip_prefix"
            AND av.binary_address STARTS WITH $prefix_binary
            AND av.prefixlen > $maxprefixlen
            AND av.version = $ip_version
            AND all(r IN relationships(path2) WHERE (%(branch_filter)s) and r.status = "active")
        WITH
            collect([pfx, av]) as all_prefixes_and_value,
            collect(pfx) as all_prefixes
        // ---
        // FIND ALL CHILDREN OF THESE PREFIXES
        // ---
        CALL (all_prefixes) {
            UNWIND all_prefixes as prefix
            OPTIONAL MATCH (prefix)<-[:IS_RELATED]-(ch_rel:Relationship)<-[:IS_RELATED]-(children:BuiltinIPPrefix)
            WHERE ch_rel.name = "parent__child"
            RETURN children
        }
        WITH collect( distinct children ) AS all_children, all_prefixes_and_value
        UNWIND all_prefixes_and_value as prefixes_to_check
        WITH prefixes_to_check, all_children
        WHERE not prefixes_to_check[0] in all_children
        """ % {
            "ns_label": InfrahubKind.IPNAMESPACE,
            "node_label": InfrahubKind.IPPREFIX,
            "branch_filter": branch_filter,
        }

        self.add_to_query(query)
        self.return_labels = ["prefixes_to_check[0] as pfx", "prefixes_to_check[1] as av"]
        self.order_by = ["av.binary_address"]

    def get_subnets(self) -> list[IPPrefixData]:
        """Return a list of all subnets fitting in the prefix."""
        subnets: list[IPPrefixData] = []

        for result in self.get_results():
            subnet = IPPrefixData(
                id=result.get("pfx").get("uuid"), prefix=ipaddress.ip_network(result.get("av").get("value"))
            )
            subnets.append(subnet)

        return subnets


class IPPrefixSubnetFetchFree(Query):
    name = "ipprefix_subnet_fetch_free"
    type = QueryType.READ
    raise_error_if_empty = False

    def __init__(
        self,
        obj: IPNetworkType,
        target_prefixlen: int,
        namespace: Node | str | None = None,
        **kwargs,
    ) -> None:
        self.obj = obj
        self.target_prefixlen = target_prefixlen
        self.namespace_id = _get_namespace_id(namespace)

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["ns_id"] = self.namespace_id

        prefix_bin = convert_ip_to_binary_str(self.obj)[: self.obj.prefixlen]
        self.params["prefix_binary"] = prefix_bin
        self.params["maxprefixlen"] = self.obj.prefixlen
        self.params["ip_version"] = self.obj.version
        self.params["parent_start"] = int(self.obj.network_address)
        self.params["parent_end"] = int(self.obj.broadcast_address)
        self.params["block_size"] = 1 << (32 - self.target_prefixlen)

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)

        # ruff: noqa: E501
        query = """
        // First match on IPNAMESPACE
        MATCH (ns:%(ns_label)s)
        WHERE ns.uuid = $ns_id
        CALL (ns) {
            MATCH (ns)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN r
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH ns, r
        WHERE r.status = "active"
        WITH ns
        OPTIONAL MATCH path2 = (ns)-[:IS_RELATED]-(ns_rel:Relationship)-[:IS_RELATED]-(pfx:%(node_label)s)-[:HAS_ATTRIBUTE]-(an:Attribute {name: "prefix"})-[:HAS_VALUE]-(av:AttributeIPNetwork)
        WHERE ns_rel.name = "ip_namespace__ip_prefix"
            AND av.binary_address STARTS WITH $prefix_binary
            AND av.prefixlen > $maxprefixlen
            AND av.version = $ip_version
            AND all(r IN relationships(path2) WHERE (%(branch_filter)s) AND r.status = "active")
        WITH collect({binary: av.binary_address, prefixlen: av.prefixlen}) AS ranges_raw
        // Convert binary strings to integer ranges for gap detection
        // - start: Convert binary string to decimal by iterating through each bit
        //          and accumulating (dec * 2 + bit_value), which is binary-to-decimal conversion
        // - end: start + (2^host_bits - 1), where host_bits = 32 - prefixlen
        //        This gives the broadcast address (last IP) of the prefix
        WITH [r IN ranges_raw |
                {
                    start: reduce(dec = 0, b IN split(r.binary, "") | dec * 2 + toInteger(b)),
                    end: reduce(dec = 0, b IN split(r.binary, "") | dec * 2 + toInteger(b)) + toInteger(2 ^ (32 - r.prefixlen)) - 1
                }
            ] AS ranges
        UNWIND CASE WHEN size(ranges) = 0 THEN [{start: null, end: null}] ELSE ranges END AS r
        WITH r
        ORDER BY r.start ASC
        WITH collect(r) AS ranges_sorted_raw
        WITH [r IN ranges_sorted_raw WHERE r.start IS NOT NULL] AS ranges_sorted
        // Gap detection algorithm using reduce to scan through sorted ranges
        // acc.cursor: current candidate position for a free block (always block-aligned)
        // acc.found: set when a gap is found, stops further iteration
        WITH reduce(acc = {cursor: $parent_start, found: null}, r IN ranges_sorted |
                CASE
                    // Already found a gap, preserve result
                    WHEN acc.found IS NOT NULL THEN acc
                    // Range ends before cursor, skip (range already passed)
                    WHEN r.end < acc.cursor THEN acc
                    // Gap found: cursor + block_size - 1 < range start means there's room before this range
                    WHEN acc.cursor + $block_size - 1 < r.start THEN {cursor: acc.cursor, found: acc.cursor}
                    // No gap: advance cursor past this range using ceiling division
                    // Formula: ceil((r.end + 1) / block_size) * block_size
                    // This aligns the cursor to the next block boundary after the range
                    ELSE
                        {
                            cursor: CASE
                                WHEN toInteger((r.end + 1 + $block_size - 1) / $block_size) * $block_size > $parent_end + 1
                                THEN $parent_end + 1
                                ELSE toInteger((r.end + 1 + $block_size - 1) / $block_size) * $block_size
                            END,
                            found: null
                        }
                END
            ) AS res
        WITH CASE
            WHEN res.found IS NOT NULL THEN res.found
            WHEN res.cursor + $block_size - 1 <= $parent_end THEN res.cursor
            ELSE NULL
        END AS free_start
        WHERE free_start IS NOT NULL
        """ % {
            "ns_label": InfrahubKind.IPNAMESPACE,
            "node_label": InfrahubKind.IPPREFIX,
            "branch_filter": branch_filter,
        }

        self.add_to_query(query)
        self.return_labels = ["free_start"]
        self.limit = 1

    def get_prefix_data(self) -> IPPrefixFreeData | None:
        result = self.get_result()
        if not result:
            return None

        return IPPrefixFreeData.from_db(result=result)


class IPv6PrefixSubnetFetchFree(Query):
    """Query to find the next free IPv6 prefix within a parent prefix.

    This query uses binary string operations to handle IPv6's 128-bit address space,
    as the integer values would overflow Neo4j's 64-bit integer type.
    """

    name = "ipv6prefix_subnet_fetch_free"
    type = QueryType.READ
    raise_error_if_empty = False

    def __init__(
        self,
        obj: IPNetworkType,
        target_prefixlen: int,
        namespace: Node | str | None = None,
        **kwargs,
    ) -> None:
        self.obj = obj
        self.target_prefixlen = target_prefixlen
        self.namespace_id = _get_namespace_id(namespace)

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["ns_id"] = self.namespace_id

        prefix_bin = convert_ip_to_binary_str(self.obj)[: self.obj.prefixlen]
        self.params["prefix_binary"] = prefix_bin
        self.params["maxprefixlen"] = self.obj.prefixlen
        self.params["ip_version"] = self.obj.version
        self.params["target_prefixlen"] = self.target_prefixlen
        # Binary representation of parent network and broadcast addresses
        self.params["parent_start_bin"] = convert_ip_to_binary_str(self.obj)
        self.params["parent_end_bin"] = format(int(self.obj.broadcast_address), "0128b")

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)

        # ruff: noqa: E501
        query = """
        // First match on IPNAMESPACE
        MATCH (ns:%(ns_label)s)
        WHERE ns.uuid = $ns_id
        CALL (ns) {
            MATCH (ns)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN r
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH ns, r
        WHERE r.status = "active"
        WITH ns
        OPTIONAL MATCH path2 = (ns)-[:IS_RELATED]-(ns_rel:Relationship)-[:IS_RELATED]-(pfx:%(node_label)s)-[:HAS_ATTRIBUTE]-(an:Attribute {name: "prefix"})-[:HAS_VALUE]-(av:AttributeIPNetwork)
        WHERE ns_rel.name = "ip_namespace__ip_prefix"
            AND av.binary_address STARTS WITH $prefix_binary
            AND av.prefixlen > $maxprefixlen
            AND av.version = $ip_version
            AND all(r IN relationships(path2) WHERE (%(branch_filter)s) AND r.status = "active")
        WITH collect({binary: av.binary_address, prefixlen: av.prefixlen}) AS ranges_raw
        // Create ranges with start (block-aligned) and end_block (the block containing the end address)
        WITH [r IN ranges_raw WHERE r.binary IS NOT NULL |
                {
                    // Start block: first target_prefixlen bits of network address
                    start_block: left(r.binary, $target_prefixlen),
                    // End block: first target_prefixlen bits of broadcast address
                    // For a prefix, broadcast = network with all host bits set to 1
                    end_block: left(left(r.binary, r.prefixlen) + reduce(s = "", i IN range(1, 128 - r.prefixlen) | s + "1"), $target_prefixlen)
                }
            ] AS ranges
        UNWIND CASE WHEN size(ranges) = 0 THEN [{start_block: null, end_block: null}] ELSE ranges END AS r
        WITH r
        ORDER BY r.start_block ASC
        WITH collect(r) AS ranges_sorted_raw
        WITH [r IN ranges_sorted_raw WHERE r.start_block IS NOT NULL] AS ranges_sorted
        // Find first available slot using binary string comparison
        // We track cursor as the current candidate block (target_prefixlen bits)
        WITH reduce(acc = {cursor: left($parent_start_bin, $target_prefixlen), found: null}, r IN ranges_sorted |
                CASE
                    WHEN acc.found IS NOT NULL THEN acc
                    // Range ends before cursor, skip it
                    WHEN r.end_block < acc.cursor THEN acc
                    // Gap found: cursor is before this range starts
                    WHEN acc.cursor < r.start_block THEN {cursor: acc.cursor, found: acc.cursor}
                    // Cursor overlaps with range: advance cursor past this range
                    // Binary string increment: add 1 to end_block to get the next block
                    // Algorithm: process bits right-to-left with carry propagation
                    // - Start with carry=1 (we're adding 1)
                    // - For each bit: if carry=0, keep bit unchanged
                    //   if carry=1 and bit="1", output "0" and carry remains 1
                    //   if carry=1 and bit="0", output "1" and carry becomes 0
                    // Note: if all bits are "1", result will be all "0"s with carry=1 (overflow)
                    // We track the final carry and set cursor to null if overflow occurred
                    ELSE
                        {
                            cursor: CASE
                                // Check for overflow: if final carry is 1, return null
                                WHEN reduce(
                                    inc = {bits: split(r.end_block, ""), carry: 1, result: []},
                                    idx IN reverse(range(0, $target_prefixlen - 1)) |
                                    {
                                        bits: inc.bits,
                                        carry: CASE
                                            WHEN inc.carry = 0 THEN 0
                                            WHEN inc.bits[idx] = "1" THEN 1
                                            ELSE 0
                                        END,
                                        result: CASE
                                            WHEN inc.carry = 0 THEN [inc.bits[idx]] + inc.result
                                            WHEN inc.bits[idx] = "1" THEN ["0"] + inc.result
                                            ELSE ["1"] + inc.result
                                        END
                                    }
                                ).carry = 1 THEN null
                                ELSE reduce(s = "", c IN reduce(
                                    inc = {bits: split(r.end_block, ""), carry: 1, result: []},
                                    idx IN reverse(range(0, $target_prefixlen - 1)) |
                                    {
                                        bits: inc.bits,
                                        carry: CASE
                                            WHEN inc.carry = 0 THEN 0
                                            WHEN inc.bits[idx] = "1" THEN 1
                                            ELSE 0
                                        END,
                                        result: CASE
                                            WHEN inc.carry = 0 THEN [inc.bits[idx]] + inc.result
                                            WHEN inc.bits[idx] = "1" THEN ["0"] + inc.result
                                            ELSE ["1"] + inc.result
                                        END
                                    }
                                ).result | s + c)
                            END,
                            found: null
                        }
                END
            ) AS res
        // Handle overflow case
        WITH
            CASE
                WHEN res.found IS NOT NULL THEN res.found
                ELSE res.cursor
            END AS cursor_str,
            res.found AS found
        // Check if we found a slot or if there's space after all ranges
        WITH CASE
            WHEN found IS NOT NULL THEN found
            // Check cursor is valid (not overflowed) and within parent range
            WHEN size(cursor_str) = $target_prefixlen AND cursor_str <= left($parent_end_bin, $target_prefixlen) THEN cursor_str
            ELSE NULL
        END AS free_start_partial
        WHERE free_start_partial IS NOT NULL
        // Pad the partial binary to full 128 bits
        WITH free_start_partial + reduce(s = "", i IN range(1, 128 - $target_prefixlen) | s + "0") AS free_start_bin
        """ % {
            "ns_label": InfrahubKind.IPNAMESPACE,
            "node_label": InfrahubKind.IPPREFIX,
            "branch_filter": branch_filter,
        }

        self.add_to_query(query)
        self.return_labels = ["free_start_bin"]
        self.limit = 1

    def get_prefix_data(self) -> IPv6PrefixFreeData | None:
        result = self.get_result()
        if not result:
            return None

        return IPv6PrefixFreeData.from_db(result=result)


class IPPrefixIPAddressFetch(Query):
    name = "ipprefix_ipaddress_fetch"
    type = QueryType.READ

    def __init__(
        self,
        obj: IPNetworkType,
        namespace: Node | str | None = None,
        **kwargs,
    ) -> None:
        self.obj = obj
        self.namespace_id = _get_namespace_id(namespace)

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["ns_id"] = self.namespace_id

        prefix_bin = convert_ip_to_binary_str(self.obj)[: self.obj.prefixlen]
        self.params["prefix_binary"] = prefix_bin
        self.params["maxprefixlen"] = self.obj.prefixlen
        self.params["ip_version"] = self.obj.version

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)

        # ruff: noqa: E501
        query = """
        // First match on IPNAMESPACE
        MATCH (ns:%(ns_label)s)
        WHERE ns.uuid = $ns_id
        CALL (ns) {
            MATCH (ns)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN r
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH ns, r
        WHERE r.status = "active"
        WITH ns
        // MATCH all IPAddress that are IN SCOPE
        MATCH path2 = (ns)-[:IS_RELATED]-(ns_rel:Relationship)-[:IS_RELATED]-(addr:%(node_label)s)-[:HAS_ATTRIBUTE]-(an:Attribute {name: "address"})-[:HAS_VALUE]-(av:AttributeIPHost)
        WHERE ns_rel.name = "ip_namespace__ip_address"
            AND av.binary_address STARTS WITH $prefix_binary
            AND av.prefixlen >= $maxprefixlen
            AND av.version = $ip_version
            AND all(r IN relationships(path2) WHERE (%(branch_filter)s) and r.status = "active")
        """ % {
            "ns_label": InfrahubKind.IPNAMESPACE,
            "node_label": InfrahubKind.IPADDRESS,
            "branch_filter": branch_filter,
        }

        self.add_to_query(query)
        self.return_labels = ["addr", "av"]
        self.order_by = ["av.binary_address"]

    def get_addresses(self) -> list[IPAddressData]:
        """Return a list of all addresses fitting in the prefix."""
        addresses: list[IPAddressData] = []

        for result in self.get_results():
            address = IPAddressData(
                id=result.get("addr").get("uuid"), address=ipaddress.ip_interface(result.get("av").get("value"))
            )
            addresses.append(address)

        return addresses


class IPPrefixIPAddressFetchFree(Query):
    name = "ipprefix_ipaddress_fetch_free"
    type = QueryType.READ

    def __init__(
        self,
        obj: IPNetworkType,
        is_pool: bool,
        namespace: Node | str | None = None,
        **kwargs,
    ) -> None:
        self.obj = obj
        self.namespace_id = _get_namespace_id(namespace)
        self.is_pool = is_pool

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["ns_id"] = self.namespace_id

        prefix_bin = convert_ip_to_binary_str(self.obj)[: self.obj.prefixlen]
        self.params["prefix_binary"] = prefix_bin
        self.params["start_range"] = int(self.obj.network_address)
        self.params["end_range"] = int(self.obj.broadcast_address)
        if not self.is_pool:
            self.params["start_range"] += 1
            self.params["end_range"] -= 1

        self.params["maxprefixlen"] = self.obj.prefixlen
        self.params["ip_version"] = self.obj.version
        self.limit = 1  # Query only works at returning a single, free entry

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)

        # ruff: noqa: E501
        query = """
        // First match on IPNAMESPACE
        MATCH (ns:%(ns_label)s)
        WHERE ns.uuid = $ns_id
        CALL (ns) {
            MATCH (ns)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN r
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH ns, r
        WHERE r.status = "active"
        WITH ns
        // MATCH all IPAddress that are IN SCOPE
        MATCH path2 = (ns)-[:IS_RELATED]-(ns_rel:Relationship)-[:IS_RELATED]-(addr:%(node_label)s)-[:HAS_ATTRIBUTE]-(an:Attribute {name: "address"})-[:HAS_VALUE]-(av:AttributeIPHost)
        WHERE ns_rel.name = "ip_namespace__ip_address"
            AND av.binary_address STARTS WITH $prefix_binary
            AND av.prefixlen >= $maxprefixlen
            AND av.version = $ip_version
            AND all(r IN relationships(path2) WHERE (%(branch_filter)s) and r.status = "active")
        ORDER BY av.binary_address
        // Gap detection algorithm: collect used addresses and find first available slot
        // Each used_address is a single binary string representing an allocated IP
        WITH DISTINCT av.binary_address AS used_address
        // Convert binary string to integer for comparison
        WITH [x IN split(used_address, "") | toInteger(x)] AS bits
        // Build array: [start_range - 1] prepended to all used addresses as integers
        // This creates a baseline for gap detection starting from the range beginning
        WITH [$start_range - 1] + collect(reduce(dec = 0, b IN bits | dec * 2 + b)) AS nums
        UNWIND range(0, size(nums) - 1) AS idx
        CALL (nums, idx) {
            // Compare expected sequential address with actual address at this position
            // If they differ, we found a gap (is_free = true)
            WITH nums[idx] AS curr, idx - 1 + $start_range AS expected
            RETURN expected AS addr, expected <> curr AS is_free, idx = size(nums) - 1 AS is_last
        }
        WITH addr, is_free, is_last
        WHERE is_free = true OR is_last = true
        WITH addr AS free_addr, is_free, is_last
        """ % {
            "ns_label": InfrahubKind.IPNAMESPACE,
            "node_label": InfrahubKind.IPADDRESS,
            "branch_filter": branch_filter,
        }

        self.add_to_query(query)
        self.return_labels = ["free_addr", "is_free", "is_last"]
        self.order_by = ["free_addr"]

    def get_address_data(self) -> IPAddressFreeData | None:
        if not self.results:
            return None

        return IPAddressFreeData.from_db(result=self.results[0])

    def get_address(self) -> IPAddressType | None:
        """Return the next free address fitting in the prefix."""
        result_data = self.get_address_data()
        if result_data is None:
            return None

        if result_data.is_free:
            return ipaddress.ip_interface(result_data.free_addr)
        # When is_last=True and is_free=False, we've reached the end of all allocated addresses
        # without finding a gap. The next available address is one past the last allocated address,
        # but only if it doesn't exceed the end of the valid range (end_range).
        if result_data.is_last and result_data.free_addr < self.params["end_range"]:
            return ipaddress.ip_interface(result_data.free_addr + 1)

        return None


class IPv6PrefixIPAddressFetchFree(Query):
    """Query to find the next free IPv6 address within a prefix.

    This query uses binary string operations to handle IPv6's 128-bit address space,
    as the integer values would overflow Neo4j's 64-bit integer type.
    """

    name = "ipv6prefix_ipaddress_fetch_free"
    type = QueryType.READ

    def __init__(
        self,
        obj: IPNetworkType,
        is_pool: bool,
        namespace: Node | str | None = None,
        **kwargs,
    ) -> None:
        self.obj = obj
        self.namespace_id = _get_namespace_id(namespace)
        self.is_pool = is_pool

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["ns_id"] = self.namespace_id

        prefix_bin = convert_ip_to_binary_str(self.obj)[: self.obj.prefixlen]
        self.params["prefix_binary"] = prefix_bin
        self.params["maxprefixlen"] = self.obj.prefixlen
        self.params["ip_version"] = self.obj.version

        # Binary representation of start and end of valid range
        start_addr = self.obj.network_address
        end_addr = self.obj.broadcast_address
        if not self.is_pool:
            start_addr += 1
            end_addr -= 1
        self.params["start_range_bin"] = format(int(start_addr), "0128b")
        self.params["end_range_bin"] = format(int(end_addr), "0128b")

        self.limit = 1  # Query only works at returning a single, free entry

        branch_filter, branch_params = self.branch.get_query_filter_path(
            at=self.at.to_string(), branch_agnostic=self.branch_agnostic
        )
        self.params.update(branch_params)

        # ruff: noqa: E501
        query = """
        // First match on IPNAMESPACE
        MATCH (ns:%(ns_label)s)
        WHERE ns.uuid = $ns_id
        CALL (ns) {
            MATCH (ns)-[r:IS_PART_OF]-(root:Root)
            WHERE %(branch_filter)s
            RETURN r
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH ns, r
        WHERE r.status = "active"
        WITH ns
        // MATCH all IPAddress that are IN SCOPE
        OPTIONAL MATCH path2 = (ns)-[:IS_RELATED]-(ns_rel:Relationship)-[:IS_RELATED]-(addr:%(node_label)s)-[:HAS_ATTRIBUTE]-(an:Attribute {name: "address"})-[:HAS_VALUE]-(av:AttributeIPHost)
        WHERE ns_rel.name = "ip_namespace__ip_address"
            AND av.binary_address STARTS WITH $prefix_binary
            AND av.prefixlen >= $maxprefixlen
            AND av.version = $ip_version
            AND all(r IN relationships(path2) WHERE (%(branch_filter)s) and r.status = "active")
        WITH DISTINCT av.binary_address AS used_address
        ORDER BY used_address
        // Collect used addresses as binary strings, prepend a sentinel value before start_range
        WITH collect(used_address) AS used_addrs_raw
        WITH [addr IN used_addrs_raw WHERE addr IS NOT NULL AND addr >= $start_range_bin AND addr <= $end_range_bin] AS used_addrs
        // Gap detection using binary string comparison
        // We iterate through used addresses and find the first gap
        WITH reduce(acc = {cursor: $start_range_bin, found: null, is_last: false}, addr IN used_addrs |
                CASE
                    // Already found a gap, preserve result
                    WHEN acc.found IS NOT NULL THEN acc
                    // Gap found: cursor is less than this address
                    WHEN acc.cursor < addr THEN {cursor: acc.cursor, found: acc.cursor, is_last: false}
                    // No gap: advance cursor past this address using binary increment
                    ELSE
                        {
                            cursor: CASE
                                // Check for overflow: if final carry is 1, return a marker value
                                WHEN reduce(
                                    inc = {bits: split(addr, ""), carry: 1, result: []},
                                    idx IN reverse(range(0, 127)) |
                                    {
                                        bits: inc.bits,
                                        carry: CASE
                                            WHEN inc.carry = 0 THEN 0
                                            WHEN inc.bits[idx] = "1" THEN 1
                                            ELSE 0
                                        END,
                                        result: CASE
                                            WHEN inc.carry = 0 THEN [inc.bits[idx]] + inc.result
                                            WHEN inc.bits[idx] = "1" THEN ["0"] + inc.result
                                            ELSE ["1"] + inc.result
                                        END
                                    }
                                ).carry = 1 THEN null
                                ELSE reduce(s = "", c IN reduce(
                                    inc = {bits: split(addr, ""), carry: 1, result: []},
                                    idx IN reverse(range(0, 127)) |
                                    {
                                        bits: inc.bits,
                                        carry: CASE
                                            WHEN inc.carry = 0 THEN 0
                                            WHEN inc.bits[idx] = "1" THEN 1
                                            ELSE 0
                                        END,
                                        result: CASE
                                            WHEN inc.carry = 0 THEN [inc.bits[idx]] + inc.result
                                            WHEN inc.bits[idx] = "1" THEN ["0"] + inc.result
                                            ELSE ["1"] + inc.result
                                        END
                                    }
                                ).result | s + c)
                            END,
                            found: null,
                            is_last: false
                        }
                END
            ) AS res
        // Determine final result
        WITH CASE
            // Gap was found during iteration
            WHEN res.found IS NOT NULL THEN {addr: res.found, is_free: true, is_last: false}
            // No gap found, check if cursor is still valid and within range
            WHEN res.cursor IS NOT NULL AND res.cursor <= $end_range_bin THEN {addr: res.cursor, is_free: true, is_last: true}
            // No addresses available
            ELSE {addr: null, is_free: false, is_last: true}
        END AS result
        WHERE result.addr IS NOT NULL
        WITH result.addr AS free_addr_bin, result.is_free AS is_free, result.is_last AS is_last
        """ % {
            "ns_label": InfrahubKind.IPNAMESPACE,
            "node_label": InfrahubKind.IPADDRESS,
            "branch_filter": branch_filter,
        }

        self.add_to_query(query)
        self.return_labels = ["free_addr_bin", "is_free", "is_last"]

    def get_address_data(self) -> IPv6AddressFreeData | None:
        if not self.results:
            return None

        return IPv6AddressFreeData.from_db(result=self.results[0])

    def get_address(self) -> IPAddressType | None:
        """Return the next free IPv6 address fitting in the prefix."""
        result_data = self.get_address_data()
        if result_data is None:
            return None

        if result_data.is_free:
            # Convert binary string to IPv6 address
            addr_int = int(result_data.free_addr_bin, 2)
            return ipaddress.ip_interface(ipaddress.IPv6Address(addr_int))

        return None


@dataclass(frozen=True)
class IPPrefixUtilizationResult:
    """Result from IPPrefixUtilization containing prefix child allocation data."""

    prefix_uuid: str
    """UUID of the parent prefix node."""

    child_uuid: str
    """UUID of the child node (prefix or address)."""

    child_kind: str
    """Kind/type of the child node."""

    child_labels: tuple[str, ...]
    """Labels of the child node (used to determine if IPADDRESS or IPPREFIX)."""

    ip_value: str
    """IP value (address or prefix) of the child."""

    prefixlen: int
    """Prefix length of the child IP value."""

    branch: str
    """Branch name where this allocation exists."""

    @classmethod
    def from_db(cls, result: QueryResult) -> IPPrefixUtilizationResult:
        """Convert raw QueryResult to typed dataclass."""
        pfx = result.get_node("pfx")
        child = result.get_node("child")
        av = result.get_node("av")
        return cls(
            prefix_uuid=str(pfx.get("uuid")),
            child_uuid=str(child.get("uuid")),
            child_kind=child.get("kind"),
            child_labels=tuple(child.labels),
            ip_value=av.get("value"),
            prefixlen=av.get("prefixlen"),
            branch=str(result.get("branch")),
        )


class IPPrefixUtilization(Query):
    name = "ipprefix_utilization_prefix"
    type = QueryType.READ

    def __init__(self, ip_prefixes: list[str], allocated_kinds: list[str], **kwargs) -> None:
        self.ip_prefixes = ip_prefixes
        self.allocated_kinds: list[str] = []
        self.allocated_kinds_rel: list[str] = []

        for kind in sorted(allocated_kinds):
            self.allocated_kinds.append(f'"{kind}"')
            self.allocated_kinds_rel.append(
                {InfrahubKind.IPADDRESS: '"ip_prefix__ip_address"', InfrahubKind.IPPREFIX: '"parent__child"'}[kind]
            )

        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        self.params["ids"] = [p.get_id() for p in self.ip_prefixes]
        self.params["time_at"] = self.at.to_string()

        def rel_filter(rel_name: str) -> str:
            return f"{rel_name}.from <= $time_at AND ({rel_name}.to IS NULL OR {rel_name}.to >= $time_at)"

        query = f"""
        MATCH (pfx:Node)
        WHERE pfx.uuid IN $ids
        CALL (pfx) {{
            MATCH (pfx)-[r_rel1:IS_RELATED]-(rl:Relationship)<-[r_rel2:IS_RELATED]-(child:Node)
            WHERE rl.name IN [{", ".join(self.allocated_kinds_rel)}]
            AND any(l IN labels(child) WHERE l IN [{", ".join(self.allocated_kinds)}])
            AND ({rel_filter("r_rel1")})
            AND ({rel_filter("r_rel2")})
            RETURN r_rel1, rl, r_rel2, child
        }}
        WITH pfx, r_rel1, rl, r_rel2, child
        MATCH path = (
            (pfx)-[r_1:IS_RELATED]-(rl:Relationship)-[r_2:IS_RELATED]-(child:Node)
            -[r_attr:HAS_ATTRIBUTE]->(attr:Attribute)
            -[r_attr_val:HAS_VALUE]->(av:{PREFIX_ATTRIBUTE_LABEL}|{ADDRESS_ATTRIBUTE_LABEL})
        )
        WHERE %(id_func)s(r_1) = %(id_func)s(r_rel1)
        AND %(id_func)s(r_2) = %(id_func)s(r_rel2)
        AND ({rel_filter("r_attr")})
        AND ({rel_filter("r_attr_val")})
        AND attr.name IN ["prefix", "address"]
        WITH
            path,
            pfx,
            child,
            av,
            reduce(br_lvl = 0, r in relationships(path) | br_lvl + r.branch_level) AS sum_branch_level,
            all(r in relationships(path) WHERE r.status = "active") AS is_active,
            [r_attr_val.from, r_attr.from, r_2.from, r_1.from] AS from_times,
            reduce(
                b_details = [0, null], r in relationships(path) |
                CASE WHEN r.branch_level > b_details[0] THEN [r.branch_level, r.branch] ELSE b_details END
            ) as deepest_branch_details
        ORDER BY pfx.uuid, child.uuid, av.uuid, sum_branch_level DESC, from_times[3] DESC, from_times[2] DESC, from_times[1] DESC, from_times[0] DESC
        WITH
            pfx,
            child,
            av,
            deepest_branch_details[0] AS branch_level,
            deepest_branch_details[1] AS branch,
            head(collect(is_active)) AS is_latest_active
        WHERE is_latest_active = TRUE
        """ % {
            "id_func": db.get_id_function_name(),
        }
        self.return_labels = ["pfx", "child", "av", "branch_level", "branch"]
        self.add_to_query(query)

    def get_data(self) -> list[IPPrefixUtilizationResult]:
        """Return results as typed dataclass instances.

        Returns:
            List of IPPrefixUtilizationResult containing prefix child allocation data.
        """
        return [IPPrefixUtilizationResult.from_db(result) for result in self.get_results()]


@dataclass(frozen=True)
class IPPrefixReconcileQueryResult:
    """Result from IPPrefixReconcileQuery containing IP reconciliation data."""

    ip_node_uuid: str | None
    """UUID of the IP node being reconciled, or None if not found."""

    current_parent_uuid: str | None
    """UUID of the current parent prefix, or None if no parent exists."""

    calculated_parent_uuid: str | None
    """UUID of the calculated correct parent prefix, or None if should be top-level."""

    current_children_uuids: tuple[str, ...]
    """UUIDs of current child prefixes/addresses."""

    calculated_children_uuids: tuple[str, ...]
    """UUIDs of calculated correct child prefixes/addresses."""

    @classmethod
    def from_db(cls, result: QueryResult) -> IPPrefixReconcileQueryResult:
        """Convert raw QueryResult to typed dataclass."""

        def get_optional_node_uuid(label: str) -> str | None:
            """Extract UUID from an optional node (may be None from OPTIONAL MATCH)."""
            node = result.get(label)
            return str(node.get("uuid")) if node and node.get("uuid") else None

        def get_collection_uuids(label: str) -> tuple[str, ...]:
            """Extract UUIDs from a node collection (may contain None from COLLECT)."""
            return tuple(str(n.get("uuid")) for n in result.get_node_collection(label) if n and n.get("uuid"))

        return cls(
            ip_node_uuid=get_optional_node_uuid("ip_node"),
            current_parent_uuid=get_optional_node_uuid("current_parent"),
            calculated_parent_uuid=get_optional_node_uuid("new_parent"),
            current_children_uuids=get_collection_uuids("current_children"),
            calculated_children_uuids=get_collection_uuids("new_children"),
        )


class IPPrefixReconcileQuery(Query):
    name = "ip_prefix_reconcile"
    type = QueryType.READ

    def __init__(
        self,
        ip_value: AllIPTypes,
        namespace: Node | str | None = None,
        node_uuid: str | None = None,
        **kwargs,
    ) -> None:
        self.ip_value = ip_value
        self.ip_uuid = node_uuid
        self.namespace_id = _get_namespace_id(namespace)
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string())
        self.params.update(branch_params)
        self.params["namespace_kind"] = InfrahubKind.IPNAMESPACE
        self.params["namespace_id"] = self.namespace_id
        self.params["ip_prefix_kind"] = InfrahubKind.IPPREFIX
        self.params["ip_address_kind"] = InfrahubKind.IPADDRESS
        self.params["ip_prefix_attribute_kind"] = PREFIX_ATTRIBUTE_LABEL
        self.params["ip_address_attribute_kind"] = ADDRESS_ATTRIBUTE_LABEL

        if isinstance(self.ip_value, IPAddressType):
            is_address = True
            prefixlen = self.ip_value.network.prefixlen
        else:
            is_address = False
            prefixlen = self.ip_value.prefixlen
        self.params["is_prefix"] = not is_address
        self.params["prefixlen"] = prefixlen
        prefix_bin = convert_ip_to_binary_str(self.ip_value)
        prefix_bin_host = prefix_bin[:prefixlen]
        self.params["prefix_binary_full"] = prefix_bin
        self.params["prefix_binary_host"] = prefix_bin_host
        self.params["ip_version"] = self.ip_value.version
        # possible prefix: highest possible prefix length for a match
        possible_prefix_map: dict[str, int] = {}
        start_prefixlen = prefixlen if is_address else prefixlen - 1
        for max_prefix_len in range(start_prefixlen, -1, -1):
            tmp_prefix = prefix_bin_host[:max_prefix_len]
            possible_prefix = tmp_prefix.ljust(self.ip_value.max_prefixlen, "0")
            if possible_prefix not in possible_prefix_map:
                possible_prefix_map[possible_prefix] = max_prefix_len
        self.params["possible_prefix_and_length_list"] = []
        self.params["possible_prefix_list"] = []
        for possible_prefix, max_length in possible_prefix_map.items():
            self.params["possible_prefix_and_length_list"].append([possible_prefix, max_length])
            self.params["possible_prefix_list"].append(possible_prefix)

        namespace_query = """
        // ------------------
        // Get IP Namespace
        // ------------------
        MATCH (ip_namespace:%(namespace_kind)s {uuid: $namespace_id})-[r:IS_PART_OF]->(root:Root)
        WHERE %(branch_filter)s
        ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
        LIMIT 1
        WITH ip_namespace
        WHERE r.status = "active"
        """ % {"branch_filter": branch_filter, "namespace_kind": self.params["namespace_kind"]}
        self.add_to_query(namespace_query)

        if self.ip_uuid:
            self.params["node_uuid"] = self.ip_uuid
            get_node_by_id_query = """
            // ------------------
            // Get IP Prefix node by UUID
            // ------------------
            OPTIONAL MATCH (ip_node:%(ip_kind)s {uuid: $node_uuid})-[r:IS_PART_OF]->(:Root)
            WHERE %(branch_filter)s
            ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
            LIMIT 1
            WITH ip_namespace, ip_node
            """ % {
                "branch_filter": branch_filter,
                "ip_kind": InfrahubKind.IPADDRESS
                if isinstance(self.ip_value, IPAddressType)
                else InfrahubKind.IPPREFIX,
            }
            self.add_to_query(get_node_by_id_query)

        else:
            get_node_by_prefix_query = """
            // ------------------
            // Get IP node with the correct value on this branch
            // ------------------
            OPTIONAL MATCH (:Root)<-[r1:IS_PART_OF]-(ip_node:%(ip_kind)s)
                -[r2:HAS_ATTRIBUTE]->(a:Attribute)-[r3:HAS_VALUE]->(aipn:%(ip_attribute_kind)s)
            WHERE aipn.binary_address = $prefix_binary_full
            AND aipn.prefixlen = $prefixlen
            AND aipn.version = $ip_version
            AND all(r IN [r1, r2, r3] WHERE (%(branch_filter)s))
            ORDER BY r3.branch_level DESC, r3.from DESC, r3.status ASC,
                r2.branch_level DESC, r2.from DESC, r2.status ASC,
                r1.branch_level DESC, r1.from DESC, r1.status ASC
            LIMIT 1
            WITH ip_namespace, CASE
                WHEN ip_node IS NOT NULL AND r1.status = "active" AND r2.status = "active" AND r3.status = "active" THEN ip_node
                ELSE NULL
            END AS ip_node
            // ------------------
            // Filter to only those that are in the correct namespace
            // ------------------
            OPTIONAL MATCH (ip_namespace)-[r1:IS_RELATED]-(nsr:Relationship)-[r2:IS_RELATED]-(ip_node)
            WHERE nsr.name IN ["ip_namespace__ip_prefix", "ip_namespace__ip_address"]
            AND all(r IN [r1, r2] WHERE (%(branch_filter)s))
            ORDER BY r1.branch_level DESC, r1.from DESC, r1.status ASC,
                r2.branch_level DESC, r2.from DESC, r2.status ASC
            LIMIT 1
            WITH ip_namespace, CASE
                WHEN ip_node IS NOT NULL AND r1.status = "active" AND r2.status = "active" THEN ip_node
                ELSE NULL
            END as ip_node
            """ % {
                "branch_filter": branch_filter,
                "ip_kind": InfrahubKind.IPADDRESS
                if isinstance(self.ip_value, IPAddressType)
                else InfrahubKind.IPPREFIX,
                "ip_attribute_kind": ADDRESS_ATTRIBUTE_LABEL
                if isinstance(self.ip_value, IPAddressType)
                else PREFIX_ATTRIBUTE_LABEL,
            }
            self.add_to_query(get_node_by_prefix_query)

        get_current_parent_query = """
        // ------------------
        // Get prefix node's current parent, if it exists
        // ------------------
        CALL (ip_node) {
            OPTIONAL MATCH parent_prefix_path = (ip_node)-[r1:IS_RELATED]->(:Relationship {name: "parent__child"})-[r2:IS_RELATED]->(current_parent:%(ip_prefix_kind)s)
            WHERE $is_prefix = TRUE
            AND all(r IN relationships(parent_prefix_path) WHERE (%(branch_filter)s))
            RETURN current_parent, (r1.status = "active" AND r2.status = "active") AS parent_is_active
            ORDER BY r1.branch_level DESC, r1.from DESC, r1.status ASC, r2.branch_level DESC, r2.from DESC, r2.status ASC
            LIMIT 1
        }
        WITH ip_namespace, ip_node, CASE WHEN parent_is_active THEN current_parent ELSE NULL END as prefix_parent
        CALL (ip_node) {
            OPTIONAL MATCH parent_prefix_path = (ip_node)-[r1:IS_RELATED]->(:Relationship {name: "ip_prefix__ip_address"})<-[r2:IS_RELATED]-(current_parent:%(ip_prefix_kind)s)
            WHERE $is_prefix = FALSE
            AND all(r IN relationships(parent_prefix_path) WHERE (%(branch_filter)s))
            RETURN current_parent, (r1.status = "active" AND r2.status = "active") AS parent_is_active
            ORDER BY r1.branch_level DESC, r1.from DESC, r1.status ASC, r2.branch_level DESC, r2.from DESC, r2.status ASC
            LIMIT 1
        }
        WITH ip_namespace, ip_node, prefix_parent, CASE WHEN parent_is_active THEN current_parent ELSE NULL END as address_parent
        WITH ip_namespace, ip_node, COALESCE(prefix_parent, address_parent) AS current_parent
        """ % {
            "branch_filter": branch_filter,
            "ip_prefix_kind": InfrahubKind.IPPREFIX,
        }
        self.add_to_query(get_current_parent_query)

        get_current_children_query = """
        // ------------------
        // Get prefix node's current prefix children, if any exist
        // ------------------
        CALL (ip_node) {
            OPTIONAL MATCH child_prefix_path = (ip_node:%(ip_prefix_kind)s)<-[r1:IS_RELATED]-(:Relationship {name: "parent__child"})<-[r2:IS_RELATED]-(current_prefix_child:%(ip_prefix_kind)s)
            WHERE all(r IN relationships(child_prefix_path) WHERE (%(branch_filter)s))
            WITH current_prefix_child, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY current_prefix_child.uuid, r1.branch_level DESC, r1.from DESC, r2.branch_level DESC, r2.from DESC
            RETURN current_prefix_child, head(collect(is_active)) AS prefix_child_is_active
        }
        WITH ip_namespace, ip_node, current_parent, CASE WHEN prefix_child_is_active THEN current_prefix_child ELSE NULL END as current_prefix_child
        WITH ip_namespace, ip_node, current_parent, collect(current_prefix_child) AS current_prefix_children
        // ------------------
        // Get prefix node's current address children, if any exist
        // ------------------
        CALL (ip_node) {
            OPTIONAL MATCH child_address_path = (ip_node:%(ip_prefix_kind)s)-[r1:IS_RELATED]->(:Relationship {name: "ip_prefix__ip_address"})<-[r2:IS_RELATED]-(current_address_child:%(ip_address_kind)s)
            WHERE all(r IN relationships(child_address_path) WHERE (%(branch_filter)s))
            WITH current_address_child, (r1.status = "active" AND r2.status = "active") AS is_active
            ORDER BY current_address_child.uuid, r1.branch_level DESC, r1.from DESC, r2.branch_level DESC, r2.from DESC
            RETURN current_address_child, head(collect(is_active)) AS address_child_is_active

        }
        WITH ip_namespace, ip_node, current_parent, current_prefix_children, CASE WHEN address_child_is_active THEN current_address_child ELSE NULL END as current_address_child
        WITH ip_namespace, ip_node, current_parent, current_prefix_children, collect(current_address_child) AS current_address_children
        WITH ip_namespace, ip_node, current_parent, current_prefix_children + current_address_children AS current_children
        """ % {
            "branch_filter": branch_filter,
            "ip_prefix_kind": InfrahubKind.IPPREFIX,
            "ip_address_kind": InfrahubKind.IPADDRESS,
        }
        self.add_to_query(get_current_children_query)

        get_new_parent_query = """
        // ------------------
        // Identify the correct parent, if any, for the prefix node
        // ------------------
        CALL (ip_namespace) {
            // ------------------
            // start with just the AttributeValue vertices b/c we have an index on them
            // ------------------
            OPTIONAL MATCH (av:%(ip_prefix_attribute_kind)s)
            WHERE av.version = $ip_version
            AND av.binary_address IN $possible_prefix_list
            AND any(prefix_and_length IN $possible_prefix_and_length_list WHERE av.binary_address = prefix_and_length[0] AND av.prefixlen <= prefix_and_length[1])
            // ------------------
            // now get all the possible IPPrefix nodes for these AttributeValues
            // ------------------
            OPTIONAL MATCH parent_path = (ip_namespace)-[:IS_RELATED]-(:Relationship {name: "ip_namespace__ip_prefix"})
            -[:IS_RELATED]-(maybe_new_parent:%(ip_prefix_kind)s)
            -[:HAS_ATTRIBUTE]->(:Attribute {name: "prefix"})
            -[:HAS_VALUE]->(av:AttributeValue)
            WHERE all(r IN relationships(parent_path) WHERE (%(branch_filter)s))
            RETURN DISTINCT maybe_new_parent
        }
        CALL (ip_namespace, maybe_new_parent) {
            // ------------------
            // filter to only active maybe_new_parent Nodes in the correct namespace
            // ------------------
            OPTIONAL MATCH (ip_namespace)-[r1:IS_RELATED]-(:Relationship {name: "ip_namespace__ip_prefix"})-[r2:IS_RELATED]-(maybe_new_parent)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH maybe_new_parent, r1, r2, r1.status = "active" AND r2.status = "active" AS is_active
            ORDER BY elementId(maybe_new_parent), r1.branch_level DESC, r1.from DESC, r1.status ASC, r2.branch_level DESC, r2.from DESC, r2.status ASC
            WITH maybe_new_parent, head(collect(is_active)) AS is_active
            RETURN is_active = TRUE AS parent_in_namespace
        }
        CALL (maybe_new_parent) {
            // ------------------
            // filter to only active maybe_new_parent Nodes currently linked to one of the allowed AttributeValues
            // ------------------
            OPTIONAL MATCH (maybe_new_parent)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "prefix"})-[r2:HAS_VALUE]->(av:AttributeValue)
            WHERE all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH maybe_new_parent, av, r1, r2, r1.status = "active" AND r2.status = "active" AS is_active
            ORDER BY elementId(maybe_new_parent), r1.branch_level DESC, r1.from DESC, r1.status ASC, r2.branch_level DESC, r2.from DESC, r2.status ASC
            // ------------------
            // get the latest active attribute value for the maybe_new_parent
            // ------------------
            WITH maybe_new_parent, head(collect([av, is_active])) AS av_is_active
            WITH
                av_is_active[0] AS av,
                av_is_active[1] AS is_active
            // ------------------
            // return NULL if the value is not allowed or if it is not active
            // ------------------
            WITH av, is_active, (
                av.version = $ip_version
                AND av.binary_address IN $possible_prefix_list
                AND any(prefix_and_length IN $possible_prefix_and_length_list WHERE av.binary_address = prefix_and_length[0] AND av.prefixlen <= prefix_and_length[1])
            ) AS is_allowed_value
            RETURN CASE WHEN is_active = TRUE AND is_allowed_value = TRUE THEN av ELSE NULL END AS allowed_av
        }
        // ------------------
        // set inactive maybe_new_parents to NULL
        // ------------------
        WITH ip_namespace, ip_node, current_parent, current_children, allowed_av.prefixlen AS mnp_prefixlen,
            CASE WHEN parent_in_namespace = TRUE AND allowed_av IS NOT NULL
                THEN maybe_new_parent ELSE NULL
            END AS maybe_new_parent
        WITH ip_namespace, ip_node, current_parent, current_children, maybe_new_parent, mnp_prefixlen
        ORDER BY ip_node.uuid, mnp_prefixlen DESC
        WITH ip_namespace, ip_node, current_parent, current_children, head(collect(maybe_new_parent)) as new_parent
        """ % {
            "branch_filter": branch_filter,
            "ip_prefix_kind": InfrahubKind.IPPREFIX,
            "ip_prefix_attribute_kind": PREFIX_ATTRIBUTE_LABEL,
        }
        self.add_to_query(get_new_parent_query)

        get_new_children_query = """
        // ------------------
        // Identify the correct children, if any, for the prefix node
        // ------------------
        CALL (ip_namespace, ip_node) {
            // ------------------
            // Get ALL possible children for the prefix node
            // ------------------
            OPTIONAL MATCH (
                 (ip_namespace)-[:IS_RELATED]-(ns_rel:Relationship)-[:IS_RELATED]
                 -(maybe_new_child:%(ip_prefix_kind)s|%(ip_address_kind)s)-[:HAS_ATTRIBUTE]
                 ->(a:Attribute)-[:HAS_VALUE]->(av:%(ip_prefix_attribute_kind)s|%(ip_address_attribute_kind)s)
            )
            USING INDEX av:%(ip_prefix_attribute_kind)s(binary_address)
            USING INDEX av:%(ip_address_attribute_kind)s(binary_address)
            WHERE $is_prefix  // only prefix nodes can have children
            AND ns_rel.name IN ["ip_namespace__ip_prefix", "ip_namespace__ip_address"]
            AND a.name in ["prefix", "address"]
            AND (ip_node IS NULL OR maybe_new_child.uuid <> ip_node.uuid)
            AND (
                ($ip_prefix_kind IN labels(maybe_new_child) AND av.prefixlen > $prefixlen)
                OR ($ip_address_kind IN labels(maybe_new_child) AND av.prefixlen >= $prefixlen)
            )
            AND av.version = $ip_version
            AND av.binary_address STARTS WITH $prefix_binary_host
            RETURN DISTINCT maybe_new_child
        }
        CALL (ip_namespace, maybe_new_child) {
            // ------------------
            // filter to only active maybe_new_child Nodes in the correct namespace
            // ------------------
            OPTIONAL MATCH (ip_namespace)-[r1:IS_RELATED]-(ns_rel:Relationship)-[r2:IS_RELATED]-(maybe_new_child)
            WHERE ns_rel.name IN ["ip_namespace__ip_prefix", "ip_namespace__ip_address"]
            AND all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH maybe_new_child, r1, r2, r1.status = "active" AND r2.status = "active" AS is_active
            ORDER BY elementId(maybe_new_child), r1.branch_level DESC, r1.from DESC, r1.status ASC, r2.branch_level DESC, r2.from DESC, r2.status ASC
            WITH maybe_new_child, head(collect(is_active)) AS is_active
            RETURN is_active = TRUE AS child_in_namespace
        }
        CALL (maybe_new_child) {
            // ------------------
            // filter to only active maybe_new_child Nodes currently linked to a possible child AttributeValue
            // ------------------
            OPTIONAL MATCH (maybe_new_child:%(ip_prefix_kind)s|%(ip_address_kind)s)-[r1:HAS_ATTRIBUTE]->(a:Attribute)-[r2:HAS_VALUE]->(av:AttributeValue)
            WHERE a.name in ["prefix", "address"]
            AND all(r IN [r1, r2] WHERE (%(branch_filter)s))
            WITH maybe_new_child, av, r1, r2, r1.status = "active" AND r2.status = "active" AS is_active
            ORDER BY elementId(maybe_new_child), r1.branch_level DESC, r1.from DESC, r1.status ASC, r2.branch_level DESC, r2.from DESC, r2.status ASC
            // ------------------
            // get the latest active attribute value for the maybe_new_child
            // ------------------
            WITH maybe_new_child, head(collect([av, is_active])) AS av_is_active
            WITH
                av_is_active[0] AS av,
                av_is_active[1] AS is_active
            // ------------------
            // return NULL if the value is not allowed or if it is not active
            // ------------------
            WITH av, is_active, (
                (
                    ($ip_prefix_kind IN labels(maybe_new_child) AND av.prefixlen > $prefixlen)
                    OR ($ip_address_kind IN labels(maybe_new_child) AND av.prefixlen >= $prefixlen)
                )
                AND av.version = $ip_version
                AND av.binary_address STARTS WITH $prefix_binary_host
            ) AS is_allowed_value
            RETURN CASE WHEN is_active = TRUE AND is_allowed_value = TRUE THEN av ELSE NULL END AS latest_mnc_attribute
        }
        // ------------------
        // set inactive/illegal value/wrong namespace maybe_new_children to NULL
        // ------------------
        WITH ip_namespace, ip_node, current_parent, current_children, new_parent,
            CASE
                WHEN child_in_namespace = TRUE AND latest_mnc_attribute IS NOT NULL THEN maybe_new_child
                ELSE NULL
            END AS maybe_new_child,
            CASE WHEN child_in_namespace = TRUE THEN latest_mnc_attribute ELSE NULL END AS latest_mnc_attribute
        WITH ip_namespace, ip_node, current_parent, current_children, new_parent, collect([maybe_new_child, latest_mnc_attribute]) AS maybe_children_ips
        WITH ip_namespace, ip_node, current_parent, current_children, new_parent, maybe_children_ips, range(0, size(maybe_children_ips) - 1) AS child_indices
        UNWIND child_indices as ind
        CALL (ind, maybe_children_ips) {
            // ------------------
            // Filter all possible children to remove those that have a more-specific parent
            // among the list of all possible children
            // ------------------
            WITH ind, maybe_children_ips AS ips
            RETURN REDUCE(
                has_more_specific_parent = FALSE, potential_parent IN ips |
                CASE
                    WHEN has_more_specific_parent THEN has_more_specific_parent  // keep it True once set
                    WHEN potential_parent IS NULL OR ips[ind][0] IS NULL THEN has_more_specific_parent
                    WHEN potential_parent[0] = ips[ind][0] THEN has_more_specific_parent  // skip comparison to self
                    WHEN $ip_address_kind in labels(potential_parent[0]) THEN has_more_specific_parent  // address cannot be a parent
                    WHEN $ip_prefix_attribute_kind IN labels(ips[ind][1]) AND (potential_parent[1]).prefixlen >= (ips[ind][1]).prefixlen THEN has_more_specific_parent  // prefix with same or greater prefixlen for prefix cannot be parent
                    WHEN $ip_address_attribute_kind IN labels(ips[ind][1]) AND (potential_parent[1]).prefixlen > (ips[ind][1]).prefixlen THEN has_more_specific_parent  // prefix with greater prefixlen for address cannot be parent
                    WHEN (ips[ind][1]).binary_address STARTS WITH SUBSTRING((potential_parent[1]).binary_address, 0, (potential_parent[1]).prefixlen) THEN TRUE  // we found a parent
                    ELSE has_more_specific_parent
                END
            ) as has_parent_among_maybe_children
        }
        WITH ip_namespace, ip_node, current_parent, current_children, new_parent, maybe_children_ips[ind][0] AS new_child, has_parent_among_maybe_children
        WHERE has_parent_among_maybe_children = FALSE
        WITH
            ip_namespace,
            ip_node,
            current_parent,
            current_children,
            new_parent,
            collect(new_child) as new_children
        """ % {
            "ip_prefix_kind": InfrahubKind.IPPREFIX,
            "ip_address_kind": InfrahubKind.IPADDRESS,
            "branch_filter": branch_filter,
            "ip_prefix_attribute_kind": PREFIX_ATTRIBUTE_LABEL,
            "ip_address_attribute_kind": ADDRESS_ATTRIBUTE_LABEL,
        }
        self.add_to_query(get_new_children_query)
        self.order_by = ["ip_node.uuid"]
        self.return_labels = ["ip_node", "current_parent", "current_children", "new_parent", "new_children"]

    def get_data(self) -> IPPrefixReconcileQueryResult | None:
        """Return single result as typed dataclass instance.

        Returns:
            IPPrefixReconcileQueryResult containing reconciliation data,
            or None if no results found.
        """
        results = list(self.get_results())
        if not results:
            return None
        return IPPrefixReconcileQueryResult.from_db(results[0])
